"""
Networking Tests.

**Assumptions**

- In docstring examples, ``listens`` callbacks are often omitted for clarity

"""

import pytest

from autopilot.networking import Net_Node, Station, Message
import numpy as np
import zmq
import time
import multiprocessing as mp



PORTRANGE = (5000, 8000)

@pytest.fixture
def node_params():
    """
    to enforce api as well as functionality...
    """
    def _node_params(**kwargs) -> dict:
        paramdict = {
            'id': '',
            'upstream': '',
            'port': np.random.randint(*PORTRANGE),
            'listens': {},
            'instance': False,
            'upstream_ip': 'localhost',
            'router_port': None,
            'daemon': True,
            'expand_on_receive': True
        }
        paramdict.update(kwargs)
        return paramdict
    return _node_params

@pytest.fixture
def station_params():
    """
    Default params for station
    """
    def _station_params(**kwargs) -> dict:
        paramdict = {
            'id': '',
            'pusher': False,
            'push_ip': 'localhost',
            'push_port': 5001,
            'push_id': '',
            'listen_port': 5000,
            'listens': None
        }
        paramdict.update(kwargs)
        return paramdict
    return _station_params


def test_node(node_params):
    """
    :class:`.Net_Node` s can be initialized with their default parameters
    """

    id = 'init'
    node = Net_Node(**node_params(id=id))
    node.release()

def test_node_to_node(node_params):
    """
    :class:`.Net_Node` s can directly send messages to each other with ``ROUTER``/``DEALER`` pairs.

    .. code-block:: python

        >>> node_1 = Net_Node(id='a', router_port=5000)
        >>> node_2 = Net_Node(id='b', upstream='a', port=5000)
        >>> node_2.send('a', 'KEY', 'VALUE')
        >>> node_2.send('b', 'KEY', 'VALUE')
    """
    global node1_received
    global node2_received
    node1_received = False
    node2_received = False

    def l_gotit(value):
        global node1_received
        global node2_received
        if value == 'node1':
            globals()['node1_received'] = True
        elif value == 'node2':
            globals()['node2_received'] = True

    node_1_params = node_params(
        id="a",
        router_port=np.random.randint(*PORTRANGE),
        listens = {'GOTIT': l_gotit}
    )
    node_2_params = node_params(
        id='b',
        upstream='a',
        port=node_1_params['router_port'],
        listens={'GOTIT': l_gotit}
    )

    node_1 = Net_Node(**node_1_params)
    node_2 = Net_Node(**node_2_params)
    time.sleep(0.1)
    node_2.send(to='a', key='GOTIT', value='node1')
    time.sleep(0.1)
    node_1.send(to='b', key='GOTIT', value='node2')
    time.sleep(0.1)

    assert node1_received
    assert node2_received

    node_1.release()
    node_2.release()


def test_multihop(node_params, station_params):
    """
    :class:`.Message` s can be routed through multiple :class:`.Station` objects
    by using a list in the ``to`` field

    .. code-block:: python

        # send message:
        # node_1 -> station_1 -> station_2 -> station_3 -> node_3
        >>> station_1 = Station(id='station_1', listen_port=6000,
                pusher=True, push_port=6001, push_id='station_2')
        >>> station_2 = Station(id='station_2', listen_port=6001,
                pusher=True, push_port=6002, push_id='station_3',)
        >>> station_3 = Station(id='station_3', listen_port=6002)
        >>> node_1 = Net_Node(id='node_1',
                upstream='station_1', port=6000)
        >>> node_3 = Net_Node(id='node_3',
                upstream='station_3', port=6002)
        >>> node_1.send(key='KEY', value='VALUE',
                to=['station_1', 'station_2', 'station_3', 'node_3'])
    """
    # node_1 -> station_1 -> station_2 -> station_3 -> node_3

    update_lock = mp.Lock()
    n_calls = mp.Value('i', lock=True)
    n_calls.value = 0

    def gotit(self):
        print(self)
        with n_calls.get_lock():
            n_calls.value += 1
    def hello(args):
        print('hello', args)
        pass

    node_1_params = node_params(
        id='node_1',
        upstream='station_1',
        port=6000,
        listens = {'HELLO': hello, 'GOTIT': gotit}
    )
    station_1_params = station_params(
        id='station_1',
        pusher=True,
        push_port=6001,
        push_id='station_2',
        listen_port=6000,
        listens={'HELLO': hello, 'GOTIT': gotit}
    )
    station_2_params = station_params(
        id='station_2',
        pusher=True,
        push_port=6002,
        push_id='station_3',
        listen_port=6001,
        listens={'HELLO': hello, 'GOTIT': gotit}
    )
    station_3_params = station_params(
        id='station_3',
        pusher=False,
        listen_port=6002,
        listens={'HELLO': hello, 'GOTIT': gotit}
    )
    node_3_params = node_params(
        id='node_3',
        upstream='station_3',
        port=6002,
        listens={'HELLO': hello, 'GOTIT': gotit}
    )

    # init in reverse
    station_3 = Station(**station_3_params)
    station_3.start()
    time.sleep(0.1)
    node_3 = Net_Node(**node_3_params)
    time.sleep(0.1)
    station_2 = Station(**station_2_params)
    station_2.start()
    time.sleep(0.1)
    station_1 = Station(**station_1_params)
    station_1.start()
    time.sleep(0.1)
    node_1 = Net_Node(**node_1_params)
    time.sleep(0.1)

    # send messages from nodes to stations to open connection
    node_1.send('station_1', 'HELLO', '')
    node_3.send('station_3', 'HELLO', '')
    time.sleep(0.1)

    # try sending message from node_1 to node_3 through the stations
    node_1.send(to=['station_1', 'station_2', 'station_3', 'node_3'],
                key='GOTIT',
                value=0)

    time.sleep(0.2)

    try:
        assert n_calls.value == 1 # gotit was only called once (no other nodes acted on it)
        assert station_1.msgs_received.value == 2 # gotit + first ping
        assert station_2.msgs_received.value == 1 # gotit
        assert station_3.msgs_received.value == 3 # extra from the confirmation
        assert node_1.msgs_received == 1 # confirmation of first ping
        assert node_3.msgs_received == 2 # confirmation of first ping + gotit
    finally:
        station_1.release()
        station_2.release()
        station_3.release()
        node_1.release()
        node_3.release()










import os
import logging
import re
import multiprocessing as mp
import inspect
from pathlib import Path
from logging.handlers import RotatingFileHandler
from threading import Lock
import warnings

from autopilot import prefs

_LOGGERS = [] # type: list
"""
List of instantiated loggers, used in :func:`.init_logger` to return existing loggers without modification
"""

_INIT_LOCK = Lock() # type: Lock


def init_logger(instance=None, module_name=None, class_name=None, object_name=None) -> logging.Logger:
    """
    Initialize a logger

    Loggers are created such that...

    * There is one logger per module (eg. all gpio objects will log to hardware.gpio)
    * If the passed object has a ``name`` attribute, that name will be prefixed to its log messages in the file
    * The loglevel for the file handler and the stdout is determined by ``prefs.get('LOGLEVEL')``, and if none is provided ``WARNING`` is used by default
    * logs are rotated according to ``prefs.get('LOGSIZE')`` (in bytes) and ``prefs.get('LOGNUM')`` (number of backups of ``prefs.get('LOGSIZE')`` to cycle through)

    Logs are stored in ``prefs.get('LOGDIR')``, and are formatted like::

        "%(asctime)s - %(name)s - %(levelname)s : %(message)s"

    Loggers can be initialized either by passing an object to the first ``instance`` argument, or
    by specifying any of ``module_name`` , ``class_name`` , or ``object_name`` (at least one must be specified)
    which are combined with periods like ``module.class_name.object_name``

    Args:
        instance: The object that we are creating a logger for! if None, at least one of ``module, class_name, or object_name`` must be passed
        module_name (None, str): If no ``instance`` passed, the module name to create a logger for
        class_name (None, str): If no ``instance`` passed, the class name to create a logger for
        object_name (None, str): If no ``instance`` passed, the object name/id to create a logger for

    Returns:
        :class:`logging.logger`
    """

    # --------------------------------------------------
    # gather variables
    # --------------------------------------------------

    if instance is not None:
        # get name of module_name without prefixed autopilot
        # eg passed autopilot.hardware.gpio.Digital_In -> hardware.gpio
        # filtering leading 'autopilot' from string

        module_name = instance.__module__
        if "__main__" in module_name:
            # awkward workaround to get module name of __main__ run objects
            mod_obj = inspect.getmodule(instance)
            try:
                mod_suffix  = inspect.getmodulename(inspect.getmodule(instance).__file__)
                module_name = '.'.join([mod_obj.__package__, mod_suffix])
            except AttributeError:
                # when running interactively or from a plugin, __main__ does not have __file__
                module_name = "__main__"


        module_name = re.sub('^autopilot.', '', module_name)

        class_name = instance.__class__.__name__

        # if object is running in separate process, give it its own file
        if issubclass(instance.__class__, mp.Process):
            # it should be at least 2 (in case its first spawned in its module)
            # but otherwise nocollide
            p_num = 2
            if module_name in globals()['_LOGGERS']:
                for existing_mod in globals()['_LOGGERS']:
                    if module_name in existing_mod and re.match(r'\d$', existing_mod):
                        p_num += 1

            module_name += f"_{str(p_num).zfill(2)}"


        # get name of object if it has one
        if hasattr(instance, 'id'):
            object_name = str(instance.id)
        elif hasattr(instance, 'name'):
            object_name = str(instance.name)
        else:
            object_name = None

        # --------------------------------------------------
        # check if logger needs to be made, or exists already
        # --------------------------------------------------
    elif not any((module_name, class_name, object_name)):
        raise ValueError('Need to either give an object to create a logger for, or one of module_name, class_name, or object_name')


    # get name of logger to get
    logger_name_pieces = [v for v in (module_name, class_name, object_name) if v is not None]
    logger_name = '.'.join(logger_name_pieces)

    # trim __ from logger names, linux don't like to make things like that
    # re.sub(r"^\_\_")

    # --------------------------------------------------
    # if new logger must be made, make it, otherwise just return existing logger
    # --------------------------------------------------

    # use a lock to prevent loggers from being double-created, just to be extra careful
    with globals()['_INIT_LOCK']:

        # check if something starting with module_name already exists in loggers
        MAKE_NEW = False
        if not any([test_logger == module_name for test_logger in globals()['_LOGGERS']]):
            MAKE_NEW = True

        if MAKE_NEW:
            parent_logger = logging.getLogger(module_name)
            loglevel = getattr(logging, prefs.get('LOGLEVEL'))
            parent_logger.setLevel(loglevel)

            # make formatter that includes name
            log_formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s : %(message)s")

            ## file handler
            # base filename is the module_name + '.log
            base_filename = (Path(prefs.get('LOGDIR')) / module_name).with_suffix('.log')

            # if directory doesn't exist, try to make it
            if not base_filename.parent.exists():
                base_filename.parent.mkdir(parents=True, exist_ok=True)

            try:
                fh = RotatingFileHandler(
                    str(base_filename),
                    mode='a',
                    maxBytes=int(prefs.get('LOGSIZE')),
                    backupCount=int(prefs.get('LOGNUM'))
                )
            except PermissionError as e:
                # catch permissions errors, try to chmod our way out of it
                try:
                    for mod_file in Path(base_filename).parent.glob(f"{Path(base_filename).stem}*"):
                        os.chmod(mod_file, 0o777)
                        warnings.warn(f'Couldnt access {mod_file}, changed permissions to 0o777')

                    fh = RotatingFileHandler(
                        base_filename,
                        mode='a',
                        maxBytes=int(prefs.get('LOGSIZE')),
                        backupCount=int(prefs.get('LOGNUM'))
                    )
                except Exception as f:
                    raise PermissionError(f'Couldnt open logfile {base_filename}, and couldnt chmod our way out of it.\n'+'-'*20+f'\ngot errors:\n{e}\n\n{f}\n'+'-'*20)

            fh.setLevel(loglevel)
            fh.setFormatter(log_formatter)
            parent_logger.addHandler(fh)

            ## log creation
            globals()['_LOGGERS'].append(module_name)
            parent_logger.info(f'parent, module-level logger created: {module_name}')

        logger = logging.getLogger(logger_name)
        logger.info(f"Logger created: {logger_name}")

    return logger


"""
Module to hold module-global variables as preferences.

Upon import, prefs attempts to import a ``prefs.json`` file from the default location (see :func:`.prefs.init` ).

Prefs are then accessed with :func:`.prefs.get` and :func:`.prefs.set` functions. After initialization, if a pref if ``set``,
it is stored in the ``prefs.json`` file -- prefs are semi-durable and persist across sessions.

When attempting to get a pref that is not set, :func:`.prefs.get` will first try to find a default value (set in
:data:`._PREFS` , and if none is found return ``None`` -- accordingly
no prefs should be intentionally set to None, as it signifies that the pref is not set.

Prefs are thread- and process-safe, as they are stored and served by a :class:`multiprocessing.Manager` object.

``prefs.json`` is typically generated by running :mod:`autopilot.setup.setup_autopilot` , though you can freestyle it
if you are so daring.

The **``HARDWARE``** pref is a little special. It specifies how each of the :mod:`.hardware` components connected to the system
is configured. It is a dictionary with this general structure::

    'HARDWARE': {
        'GROUP': {
            'ID': {
                'hardware_arg': 'val'
            }
        }
    }

where there are user-named ``'GROUPS'`` of hardware objects, like ``'LEDS'`` , etc. Within a group, each object has its
``'ID'`` (passed as the ``name`` argument to the hardware initialization method) which allows it to be identified from
the other components in the group. The intention of this structure is to allow multiple categories of hardware objects
to be parameterized and used separately, even though they might be the same object type. Eg. we may have three LEDs
in our nosepokes, but also have an LED that serves at the arena light. If we wanted to write a command that turns off all
LEDs, we would have to explicitly specify their IDs, making it difficult to re-use very common hardware command patterns
within tasks. There are obvious drawbacks to this scheme -- clunky, ambiguous, etc. and will be deprecated as parameterization
continues to congeal across the library.

The class that each element is used with is determined by the :attr:`.Task.HARDWARE`
dictionary. Specifically, the :meth:`.Task.init_hardware` method does something like::

    self.hardware['GROUP']['ID'] = self.HARDWARE['GROUP']['ID'](**prefs.get('HARDWARE')['GROUP']['ID'])


Warning:
    These are **not** hard coded prefs. :data:`_DEFAULTS` populates the *default* values for prefs, but local prefs are
    always restored from and saved to ``prefs.json`` . If you're editing this file and things aren't changing,
    you're in the wrong place!

This iteration of prefs with respect to work done on the `People's Ventilator Project <https://www.peoplesvent.org/en/latest/pvp.common.prefs.html>`_

If a pref has a string for a ``'deprecation'`` field in :data:`.prefs._DEFAULTS` , a ``FutureWarning``
will be raised with the string given as the message


"""

# this is strictly a placeholder module to
# allow global access to prefs without explicit passing.
#
# DO NOT hardcode prefs here. only add placeholder values for certain 'universal' params
#
# A prefs.json file should be generated with an appropriate setup routine
# (see setup dir)
# then you should call prefs.init(prefs.json) if the if __name__=="__main__" block

# Prefs is a top-level module! It shouldn't depend on anything else in Autopilot,
# and if it does, it should carefully import it where it is needed!
# (prefs needs to be possible to import everywhere, including eg. in setup_autopilot)
import json
import subprocess
import multiprocessing as mp
import os
import logging
import typing
import sys
import types
from pathlib import Path
from ctypes import c_bool
from enum import Enum, auto
import warnings

#from autopilot.core.loggers import init_logger
from collections import OrderedDict as odict

class Scopes(Enum):
    """
    Enum that lists available scopes and groups for prefs

    Scope can be an agent type, common (for everyone), or specify some
    subgroup of prefs that should be presented together (like directories)

    COMMON = All Agents
    DIRECTORY = Prefs group for specifying directory structure
    TERMINAL = prefs for Terminal Agents
    Pilot = Prefs for Pilot agents
    LINEAGE = prefs for networking lineage (until networking becomes more elegant ;)
    AUDIO = Prefs for configuring the Jackd audio server


    """
    COMMON = auto() #: All agents
    TERMINAL = auto() #: Prefs specific to Terminal Agents
    PILOT = auto() #: Prefs specific to Pilot Agents
    DIRECTORY = auto() #: Directory structure
    LINEAGE = auto() #: Prefs for coordinating network between pilots and children
    AUDIO = auto() #: Audio prefs...




_PREF_MANAGER = mp.Manager() # type: mp.managers.SyncManager
"""
The :class:`multiprocessing.Manager` that stores prefs during system operation and makes them available
and consistent across processes.
"""

_PREFS = _PREF_MANAGER.dict() # type: mp.managers.SyncManager.dict
"""
stores a dictionary of preferences that mirrors the global variables.
"""

_LOGGER = None # type: typing.Union[logging.Logger, None]
"""
Logger used by prefs initialized by :func:`.core.loggers.init_logger`

Initially None, created once prefs are populated because init_logger requires some prefs to be set (uh the logdir and level and stuff)
"""

_INITIALIZED = mp.Value(c_bool, False) # type: mp.Value
"""
Boolean flag to indicate whether prefs have been initialzied from ``prefs.json``
"""

_LOCK = mp.Lock() # type: mp.Lock
"""
:class:`multiprocessing.Lock` to control access to ``prefs.json``
"""

# not documenting, just so that the full function doesn't need to be put in for each directory
# lol at this literal reanimated fossil halfway evolved between os.path and pathlib
_basedir = Path(os.path.join(os.path.expanduser("~"), "autopilot"))


_DEFAULTS = odict({
    'NAME': {
        'type': 'str',
        "text": "Agent Name:",
        "scope": Scopes.COMMON
    },
    'PUSHPORT': {
        'type': 'int',
        "text": "Push Port - Router port used by the Terminal or upstream agent:",
        "default": "5560",
        "scope": Scopes.COMMON
    },
    'MSGPORT': {
        'type': 'int',
        "text": "Message Port - Router port used by this agent to receive messages:",
        "default": "5565",
        "scope": Scopes.COMMON
    },
    'TERMINALIP': {
        'type': 'str',
        "text": "Terminal IP:",
        "default": "192.168.0.100",
        "scope": Scopes.COMMON
    },
    'LOGLEVEL': {
        'type': 'choice',
        "text": "Log Level:",
        "choices": ("DEBUG", "INFO", "WARNING", "ERROR"),
        "default": "WARNING",
        "scope": Scopes.COMMON
    },
    'LOGSIZE': {
        'type': 'int',
        "text": "Size of individual log file (in bytes)",
        "default": 5 * (2 ** 20),  # 50MB
        "scope": Scopes.COMMON
    },
    'LOGNUM': {
        'type': 'int',
        "text": "Number of logging backups to keep of LOGSIZE",
        "default": 4,
        "scope": Scopes.COMMON
    },
    # 4 * 5MB = 20MB per module
    'CONFIG': {
        'type': 'list',
        "text": "System Configuration",
        'hidden': True,
        "scope": Scopes.COMMON
    },
    'VENV': {
        'type': 'str',
        'text': 'Location of virtual environment, if used.',
        "scope": Scopes.COMMON,
        "default": str(Path(sys.prefix).resolve()) if hasattr(sys, 'real_prefix') or (sys.base_prefix != sys.prefix) else False
    },
    'AUTOPLUGIN': {
        'type': 'bool',
        'text': "Attempt to import the contents of the plugin directory",
        "scope": Scopes.COMMON,
        "default": True
    },
    'PLUGIN_DB': {
        'type': 'str',
        'text': 'filename to use for the .json plugin_db that keeps track of installed plugins',
        "default": str(_basedir / "plugin_db.json"),
        "scope": Scopes.COMMON
    },
    'BASEDIR': {
        'type': 'str',
        "text": "Base Directory",
        "default": str(_basedir),
        "scope": Scopes.DIRECTORY
    },
    'DATADIR': {
        'type': 'str',
        "text": "Data Directory",
        "default": str(_basedir / 'data'),
        "scope": Scopes.DIRECTORY
    },
    'SOUNDDIR': {
        'type': 'str',
        "text": "Sound file directory",
        "default": str(_basedir / 'sounds'),
        "scope": Scopes.DIRECTORY
    },
    'LOGDIR': {
        'type': 'str',
        "text": "Log Directory",
        "default": str(_basedir / 'logs'),
        "scope": Scopes.DIRECTORY
    },
    'VIZDIR': {
        'type': 'str',
        "text": "Directory to store Visualization results",
        "default": str(_basedir / 'viz'),
        "scope": Scopes.DIRECTORY
    },
    'PROTOCOLDIR': {
        'type': 'str',
        "text": "Protocol Directory",
        "default": str(_basedir / 'protocols'),
        "scope": Scopes.DIRECTORY
    },
    'PLUGINDIR': {
        'type': 'str',
        "text": "Directory to import ",
        "default": str(_basedir / 'plugins'),
        "scope": Scopes.DIRECTORY
    },
    'REPODIR': {
       'type': 'str',
       'text': 'Location of Autopilot repo/library',
        'default': Path(__file__).resolve().parents[1],
        "scope": Scopes.DIRECTORY
    },
    'CALIBRATIONDIR': {
        'type': 'str',
        'text': 'Location of calibration files for solenoids, etc.',
        'default': str(_basedir / 'calibration'),
        'scope': Scopes.DIRECTORY
    },
    'PIGPIOMASK': {
        'type': 'str',
        'text': 'Binary mask controlling which pins pigpio controls according to their BCM numbering, see the -x parameter of pigpiod',
        'default': "1111110000111111111111110000",
        "scope": Scopes.PILOT
    },
    'PIGPIOARGS': {
        'type': 'str',
        'text': 'Arguments to pass to pigpiod on startup',
        'default': '-t 0 -l',
        "scope": Scopes.PILOT
    },
    'PULLUPS': {
        'type': 'list',
        'text': 'Pins to pull up on system startup? (list of form [1, 2])',
        "scope": Scopes.PILOT
    },
    'PULLDOWNS': {
        'type': 'list',
        'text': 'Pins to pull down on system startup? (list of form [1, 2])',
        "scope": Scopes.PILOT
    },
    'PING_INTERVAL': {
        'type': 'float',
        'text': 'How many seconds should pilots wait in between pinging the Terminal?',
        'default': 5,
        'scope': Scopes.PILOT
    },
    'DRAWFPS': {
        'type': 'int',
        "text": "FPS to draw videos displayed during acquisition",
        "default": "20",
        "scope": Scopes.TERMINAL
    },
    'PILOT_DB': {
        'type': 'str',
        'text': "filename to use for the .json pilot_db that maps pilots to subjects (relative to BASEDIR)",
        "default": str(_basedir / "pilot_db.json"),
        "scope": Scopes.TERMINAL
    },
    'TERMINAL_SETTINGS_FN':{
        'type': 'str',
        'text': 'filename to store QSettings file for Terminal',
        'default': str(_basedir / "terminal.conf"),
        "scope": Scopes.TERMINAL
    },
    'TERMINAL_WINSIZE_BEHAVIOR': {
        'type': 'choice',
        'text': 'Strategy for resizing terminal window on opening',
        "choices": ('remember', 'moderate', 'maximum', 'custom'),
        "default": "remember",
        "scope": Scopes.TERMINAL    
    },
    'TERMINAL_CUSTOM_SIZE': {
        'type': 'list',
        'text': 'Custom size for window, specified as [px from left, px from top, width, height]',
        'default': [0, 0, 1000, 400],
        'depends': ('TERMINAL_WINSIZE_BEHAVIOR', 'custom'),
        'scope': Scopes.TERMINAL
    },
    'LINEAGE': {
        'type': 'choice',
        "text": "Are we a parent or a child?",
        "choices": ("NONE", "PARENT", "CHILD"),
        "scope": Scopes.LINEAGE
    },
    'CHILDID': {
        'type': 'list',
        "text": "List of Child ID:",
        'default': [],
        "depends": ("LINEAGE", "PARENT"),
        "scope": Scopes.LINEAGE
    },
    'PARENTID': {
        'type': 'str',
        "text": "Parent ID:",
        "depends": ("LINEAGE", "CHILD"),
        "scope": Scopes.LINEAGE
    },
    'PARENTIP': {
        'type': 'str',
        "text": "Parent IP:",
        "depends": ("LINEAGE", "CHILD"),
        "scope": Scopes.LINEAGE
    },
    'PARENTPORT': {
        'type': 'str',
        "text": "Parent Port:",
        "depends": ("LINEAGE", "CHILD"),
        "scope": Scopes.LINEAGE
    },
    'AUDIOSERVER': {
        'type': 'bool',
        'text': 'Enable jack audio server?',
        "scope": Scopes.AUDIO
    },
    'NCHANNELS': {
        'type': 'int',
        'text': "Number of Audio channels (deprecated; used OUTCHANNELS)",
        'default': 1,
        'depends': 'AUDIOSERVER',
        "scope": Scopes.AUDIO,
        'deprecation': "Deprecated and will be removed, use OUTCHANNELS instead"
    },
    'OUTCHANNELS': {
        'type': 'list',
        'text': 'List of Audio channel indexes to connect to',
        'default': '',
        'depends': 'AUDIOSERVER',
        "scope": Scopes.AUDIO
    },
    'FS': {
        'type': 'int',
        'text': 'Audio Sampling Rate',
        'default': 192000,
        'depends': 'AUDIOSERVER',
        "scope": Scopes.AUDIO
    },
    'JACKDSTRING': {
        'type': 'str',
        'text': 'Arguments to pass to jackd, see the jackd manpage',
        'default': 'jackd -P75 -p16 -t2000 -dalsa -dhw:sndrpihifiberry -P -rfs -n3 -s &',
        'depends': 'AUDIOSERVER',
        "scope": Scopes.AUDIO
    },
})
"""
Ordered Dictionary containing default values for prefs.

An Ordered Dictionary lets the prefs be displayed in gui elements in a predictable order, but prefs are stored in ``prefs.json`` in
alphabetical order and the 'live' prefs used during runtime are stored in :data:`._PREFS`

Each entry should be a dict with the following structure::

    "PREF_NAME": {
        "type": (str, int, bool, choice, list) # specify the appropriate GUI input, str or int are validators, 
        choices are a 
            # dropdown box, and lists allow users to specify lists of values like "[0, 1]"
        "default": If possible, assign default value, otherwise None
        "text": human-readable text that described the pref
        "scope": to whom does this pref apply? see :class:`.Scopes`
        "depends": name of another pref that needs to be supplied/enabled for this one to be enabled (eg. don't set sampling rate of audio server if audio server disabled)
            can also be specified as a tuple like ("LINEAGE", "CHILD") that enables the option when prefs[depends[0]] == depends[1]
        "choices": If type=="choice", a tuple of available choices.
    }
"""


def get(key: typing.Union[str, None] = None):
    """
    Get a pref!

    If a value for the given ``key`` can't be found, prefs will attempt to

    Args:
        key (str, None): get pref of specific ``key``, if ``None``, return all prefs

    Returns:
        value of pref (type variable!), or ``None`` if no pref of passed ``key``
    """

    # if nothing is requested of us, return everything
    if key is None:
        return globals()['_PREFS']._getvalue()

    else:
        # check for deprecation
        dep_notice = globals()['_DEFAULTS'].get(key, {}).get('deprecation', None)
        if dep_notice is not None:
            warnings.warn(dep_notice, FutureWarning)

        # try to get the value from the prefs manager
        try:
            # if it's a directory and it doesn't exist, try and make it
            if globals()['_DEFAULTS'].get(key, {}).get('scope', False) == Scopes.DIRECTORY:
                try:
                    path = Path(globals()['_PREFS'][key]).resolve()
                    if not path.exists():
                        path.mkdir(parents=True, exist_ok=True)
                except Exception as e:
                    warnings.warn(f"prefs {key} was a directory, but a directory couldnt be created. got exception {e}")
            return globals()['_PREFS'][key]

        # if none exists...
        except KeyError:
            # try to get a default value
            try:
                default_val = globals()['_DEFAULTS'][key]['default']
                warnings.warn(f'Returning default prefs value {key} : {default_val} (ideally this shouldnt happen and everything should be specified in prefs', UserWarning)
                return default_val

            # if you still can't find a value, None is an unambiguous signal for pref not set
            # (no pref is ever None)
            except KeyError:
                return None

def set(key: str, val):
    """
    Set a pref!

    Note:
        Whenever a pref is set, the prefs file is automatically updated -- prefs are system-durable!!

        (specifically, whenever the module-level ``_INITIALIZED`` value is set to True, prefs are saved to file to
        avoid overwriting before loading)

    Args:
        key (str): Name of pref to set
        val: Value of pref to set (prefs are not type validated against default types)
    """
    globals()['_PREFS'][key] = val
    if globals()['_INITIALIZED'].value and 'pytest' not in sys.modules:
        save_prefs()


def save_prefs(prefs_fn: str = None):
    """
    Dump prefs into the ``prefs_fn`` .json file

    Args:
        prefs_fn (str, None): if provided, pathname to ``prefs.json`` otherwise resolve ``prefs.json`` according the
        to the normal methods....
    """
    if prefs_fn is None:
        try:
            prefs_fn = str(Path(get('BASEDIR')) / 'prefs.json')
        except KeyError:
            raise RuntimeError('Asked to save prefs without BASEDIR being set -- indicative of prefs being saved '
                               'before initialized')

    # take lock for access to prefs file
    with globals()['_LOCK']:
        with open(prefs_fn, 'w') as prefs_f:
            json.dump(globals()['_PREFS']._getvalue(), prefs_f,
                      indent=4, separators=(',', ': '))



def init(fn=None):
    """
    Initialize prefs on autopilot start.

    If passed dict of prefs or location of prefs.json, load and use that

    Otherwise

    - Look for the autopilot wayfinder ``~/.autopilot`` file that tells us where the user directory is
    - look in default location ``~/autopilot/prefs.json``

    Todo:

        This function may be deprecated in the future -- in its current form it serves to allow the sorta janky launch
        methods in the headers/footers of autopilot/core/pilot.py and autopilot/core/terminal.py that will eventually
        be transformed into a unified agent framework to make launching easier. Ideally one would be able to just
        import prefs without having to explicitly initialize it, but we need to formalize the full launch process
        before we make the full lurch to that model.

    Args:
        fn (str, dict): a path to `prefs.json` or a dictionary of preferences
    """
    if isinstance(fn, str):
        with open(fn, 'r') as pfile:
            prefs = json.load(pfile)
    elif isinstance(fn, dict):
        prefs = fn
    elif fn is None:
        # try to load from default location
        autopilot_wayfinder = os.path.join(os.path.expanduser('~'), '.autopilot')
        if os.path.exists(autopilot_wayfinder):
            with open(autopilot_wayfinder, 'r') as wayfinder_f:
                fn = os.path.join(wayfinder_f.read(), 'prefs.json')
        else:
            fn = os.path.join(os.path.expanduser('~'), 'autopilot', 'prefs.json')

        if not os.path.exists(fn):
            # tried to load defaults, return quietly
            return

        with open(fn, 'r') as pfile:
            prefs = json.load(pfile)

    # Get the current git hash
    if prefs.get('REPODIR', False):
        try:
            prefs['HASH'] = git_version(prefs.get('REPODIR'))
        except Exception as e:
            prefs['HASH'] = ''
            warnings.warn(f'git hash for repo could not be found! will not be able to keep good provenance! got exception: \n{e}')
    else:
        warnings.warn('REPODIR is not set in prefs.json, cant get git hash!!!')

    # FIXME: This 100% should not happen here and should happen in the relevant hardware classes.
    # Load any calibration data
    if prefs.get('BASEDIR', False):
        cal_path = os.path.join(prefs['BASEDIR'], 'port_calibration_fit.json')
        cal_raw = os.path.join(prefs['BASEDIR'], 'port_calibration.json')

        if os.path.exists(cal_path):
            with open(cal_path, 'r') as calf:
                cal_fns = json.load(calf)
            prefs['PORT_CALIBRATION'] = cal_fns
        elif os.path.exists(cal_raw):
            # aka raw calibration results exist but no fit has been computed
            luts = compute_calibration(path=cal_raw, do_return=True)
            with open(cal_path, 'w') as calf:
                json.dump(luts, calf)
            prefs['PORT_CALIBRATION'] = luts

    ###########################

    global _PREFS

    for k, v in prefs.items():
        # globals()[k] = v
        _PREFS[k] = v

    # also store as a dictionary so other modules can have one if they want it
    # globals()['__dict__'] = prefs

    globals()['_INITIALIZED'].value = True

def add(param, value):
    """
    Add a pref after init

    Args:
        param (str): Allcaps parameter name
        value: Value of the pref
    """
    globals()[param] = value

    global _PREFS
    _PREFS[param] = value

# Return the git revision as a string
def git_version(repo_dir):
    """
    Get the git hash of the current commit.

    Stolen from `numpy's setup <https://github.com/numpy/numpy/blob/master/setup.py#L70-L92>`_

    and linked by ryanjdillon on `SO <https://stackoverflow.com/a/40170206>`_


    Args:
        repo_dir (str): directory of the git repository.

    Returns:
        unicode: git commit hash.
    """
    def _minimal_ext_cmd(cmd):
        # type: (list[str]) -> str
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env['LANGUAGE'] = 'C'
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        out = subprocess.Popen(cmd, stdout = subprocess.PIPE, env=env).communicate()[0]
        return out

    out = _minimal_ext_cmd(['git','-C',repo_dir, 'rev-parse', 'HEAD'])
    GIT_REVISION = out.strip().decode('ascii')

    return GIT_REVISION


def compute_calibration(path=None, calibration=None, do_return=False):
    """

    Args:
        path:
        calibration:
        do_return:

    Returns:

    """
    # FIXME: UGLY HACK - move this function to another module
    import pandas as pd
    from scipy.stats import linregress

    if not calibration:
        # if we weren't given calibration results, load them
        if path:
            open_fn = path
        else:
            open_fn = "/usr/autopilot/port_calibration.json"

        with open(open_fn, 'r') as open_f:
            calibration = json.load(open_f)

    luts = {}
    for port, samples in calibration.items():
        sample_df = pd.DataFrame(samples)
        # TODO: Filter for only most recent timestamps

        # volumes are saved in mL because of how they are measured, durations are stored in ms
        # but reward volumes are typically in the uL range, so we make the conversion
        # by multiplying by 1000
        line_fit = linregress((sample_df['vol'] / sample_df['n_clicks']) * 1000., sample_df['dur'])
        luts[port] = {'intercept': line_fit.intercept,
                      'slope': line_fit.slope}

    # write to file, overwriting any previous
    if do_return:
        return luts

    else:
        # do write
        lut_fn = os.path.join(globals()['BASEDIR'], 'port_calibration_fit.json')
        with open(lut_fn, 'w') as lutf:
            json.dump(luts, lutf)

def clear():
    """
    Mostly for use in testing, clear loaded prefs (without deleting prefs.json)

    (though you will probably overwrite prefs.json if you clear and then set another pref so don't use this except in testing probably)
    """
    global _PREFS
    global _PREF_MANAGER
    _PREFS = _PREF_MANAGER.dict()



#######################3

if not _INITIALIZED.value:
    init()

    # replace module
    # sys.modules[__name__] = _Prefs(__name__)





"""
Data transformations.

Composable transformations from one representation of data to another.
Used as the lubricant and glue between hardware objects. Some hardware objects
disagree about the way information should be represented -- eg. cameras are very
partial to letting position information remain latent in a frame of a video, but
some other object might want the actual ``[x,y]`` coordinates. Transformations help
negotiate (but don't resolve their irreparably different worldviews :( )

Transformations are organized by modality, but this API is quite immature.

Transformations have a ``process`` method that accepts and returns a single object.
They must also define the format of their inputs and outputs (``format_in``
and ``format_out``). That API is also a sketch.

The :meth:`~.Transform.__add__` method allows transforms to be combined, eg.::

    from autopilot import transform as t
    transform_me = t.Image.DLC('model_directory')
    transform_me += t.selection.DLCSlice('point')
    transform_me.process(frame)
    # ... etcetera

.. todo::

    This is a first draft of this module and it purely synchronous at the moment. It will be expanded to ...
    * support multiple asynchronous processing rhythms
    * support automatic value coercion
    * make recursion checks -- make sure a child hasn't already been added to a processing chain.
    * idk participate at home! list your own shortcomings of this module, don't be shy it likes it.
"""

import typing
import autopilot
from autopilot.transform.transforms import Transform
from autopilot.transform import image, geometry, logical, selection, units

IMPORTED = False


def make_transform(transforms: typing.Union[typing.List[dict],typing.Tuple[dict]]) -> Transform:
    """
    Make a transform from a list of iterator specifications.

    Args:
        transforms (list): A list of :class:`Transform` s and parameterizations in the form::

            [
                {'transform': Transform,
                'args': (arg1, arg2,), # optional
                'kwargs': {'key1':'val1', ...}, # optional
                {'transform': ...}
            ]

    Returns:
        :class:`Transform`
    """

    ret = None
    for t in transforms:
        if isinstance(t['transform'], str):
            if len(t['transform'].split('.'))>1:
                module, classname = t['transform'].split('.')
                mod = __import__(f"autopilot.transform.{module}", fromlist=[classname])
                tfm_class = getattr(mod, classname)
            else:
                tfm_class = autopilot.get('transform', t['transform'])

        elif issubclass(t['transform'], Transform):
            tfm_class = t['transform']
        else:
            raise ValueError(f'Could not get transform from {t["transform"]}, need a name of a Transform class, or the class itself')

        transform = tfm_class(
            *t.get('args', []),
            **t.get('kwargs', {})
        )
        if ret is None:
            ret = transform
        else:
            ret = ret + transform

    return ret



"""
This is a scrappy first draft of a stimulus manager that will be built out
to incorporate arbitrary stimulus logic. For now you can subclass `Stim_Manager` and
redefine `next_stim`

TODO:
    Make this more general, for more than just sounds.
"""

import os
import pdb
from collections import deque
import numpy as np
import autopilot
from autopilot import prefs
if prefs.get('AGENT') and prefs.get('AGENT').upper() == 'PILOT':
    if 'AUDIO' in prefs.get('CONFIG') or prefs.get('AUDIOSERVER') is not None:
        from autopilot.stim.sound import sounds

def init_manager(stim):
    if 'manager' in stim.keys():
        if stim['manager'] in MANAGER_MAP.keys():
            manager = stim['manager']
            return MANAGER_MAP[manager](stim)
        else:
            Exception('Couldnt find stim manager of type: {}'.format(stim['type']))
    else:
        return Stim_Manager(stim)


class Stim_Manager(object):
    """
    Yield sounds according to some set of rules.

    Currently implemented:

    * correction trials - If a subject continually answers to one side incorrectly, keep
        the correct answer on the other side until they answer in that direction
    * bias correction - above some bias threshold, skew the correct answers to the less-responded side

    Attributes:
        stimuli (dict): Dictionary of instantiated stimuli like::

            {'L': [Tone1, Tone2, ...], 'R': [Tone3, Tone4, ...]}

        target ('L', 'R'): What is the correct port?
        distractor ('L', 'R'): What is the incorrect port?
        response ('L', 'R'): What was the last response?
        correct (0, 1): Was the last response correct?
        last_stim: What was the last stim? (one of `self.stimuli`)
        correction (bool): Are we doing correction trials?
        correction_trial (bool): Is this a correction trial?
        last_was_correction (bool): Was the last trial a correction trial?
        correction_pct (float): proportion of trials that are correction trials
        bias: False, or a bias correction mode.

    """

    # Metaclass for managing stimuli...
    def __init__(self, stim=None):
        """
        Args:
            stim (dict): Dictionary describing sound stimuli, in a format like::

                {
                'L': [{'type':'tone',...},{...}],
                'R': [{'type':'tone',...},{...}]
                }
        """
        self.stimuli = {}

        self.target = None  # What is the correct port?
        self.distractor = None  # What is the incorrect port
        self.response = None  # What was the last response?
        self.correct = 0  # Was the last response correct?
        self.last_stim = None  # What was the last stim?

        # Correction trials

        self.correction = False  # Are we doing correction trials
        self.correction_trial = False  # Is this a correction trial?
        self.last_was_correction = False  # Was the last trial a correction trial?
        self.correction_pct = 0.1  # proportion of trials that are correction trials

        # Bias correction
        self.bias = False  # or a bias correction mode

        # if we're being init'd as a superclass, no stim passed.
        if stim:

            # if we're doing sounds init them here
            # TODO: Make SoundManger subclass
            if 'sounds' in stim.keys():
                self.init_sounds(stim['sounds'])

    def do_correction(self, correction_pct = 0.5):
        """
        Called to set correction trials to True and correction percent.

        Args:
            correction_pct (float): Proportion of trials that should randomly be set to be correction trials.
        """
        self.correction = True
        self.correction_pct = correction_pct

    def do_bias(self, **kwargs):
        """
        Instantiate a :class:`.Bias_Correction` module

        Args:
            kwargs: parameters to initialize :class:`.Bias_Correction` with.
        """
        self.bias = Bias_Correction(**kwargs)

    def init_sounds(self, sound_dict):
        """
        Instantiate sound objects, using the 'type' value to choose an object from
        ``autopilot.get('sound')`` .

        Args:
            sound_dict (dict): a dictionary like::
                {
                'L': [{'type':'tone',...},{...}],
                'R': [{'type':'tone',...},{...}]
                }
        """
        # sounds should be
        # Iterate through sounds and load them to memory
        for k, v in sound_dict.items():
            # If multiple sounds on one side, v will be a list
            if isinstance(v, list):
                self.stimuli[k] = []
                for sound in v:
                    # We send the dict 'sound' to the function specified by 'type' and '
                    # ' as kwargs
                    self.stimuli[k].append(autopilot.get('sound', sound['type'])(**sound))
            # If not a list, a single sound
            else:
                self.stimuli[k] = [autopilot.get('sound', v['type'])(**v)]

    def set_triggers(self, trig_fn):
        """
        Give a callback function to all of our stimuli for when the stimulus ends.

        Note:
            Stimuli need a `set_trigger` method.

        Args:
            trig_fn (callable): A function to be given to stimuli via `set_trigger`
        """
        # set a callback function for when the stimulus ends
        for k, v in self.stimuli.items():
            for astim in v:
                astim.set_trigger(trig_fn)

    def make_punishment(self, type, duration):
        """
        Warning:
            Not Implemented

        Args:
            type:
            duration:
        """
        # types: timeout, noise
        # If we want a punishment sound...
        # if self.punish_sound:
        #     self.stimuli['punish'] = sounds.Noise(self.punish_dur)
        #     #change_to_green = lambda: self.hardware['LEDS']['C'].set_color([0, 255, 0])
        #     #self.stimuli['punish'].set_trigger(change_to_green)
        pass

    def play_punishment(self):
        """
        Warning:
            Not Implemented
        """
        pass

    def next_stim(self):
        """
        Compute and return the next stimulus

        If we are doing correction trials, compute that.

        Same thing with bias correction.

        Otherwise, randomly select a stimulus to present.

        Returns:
            ('L'/'R' Target, 'L'/'R' distractor, Stimulus to present)
        """
        # compute and return the next stim

        # first: if we're doing correction trials, compute that
        if self.correction:
            self.correction_trial = self.compute_correction()
            if self.correction_trial:
                return self.target, self.distractor, self.last_stim

        # otherwise we check for bias correction
        # it will return a threshold for random choice
        if self.bias:
            threshold = self.bias.next_bias()
        else:
            threshold = 0.5

        if np.random.rand()<threshold:
            self.target = 'L'
        else:
            self.target = 'R'

        if self.target == 'L':
            self.distractor = 'R'
        elif self.target == 'R':
            self.distractor = 'L'

        self.last_stim = np.random.choice(self.stimuli[self.target])

        return self.target, self.distractor, self.last_stim

    def compute_correction(self):
        """
        If `self.correction` is true, compute correction trial logic during
        `next_stim`.

        * If the last trial was a correction trial and the response to it wasn't correct, return True
        * If the last trial was a correction trial and the response was correct, return False
        * If the last trial as not a correction trial, but a randomly generated float is less than `correction_pct`, return True.

        Returns:
            bool: whether this trial should be a correction trial.

        """
        # if we are doing a correction trial this time return true

        # if this is the first trial, we can't do correction trials now can we
        if self.target is None:
            return False

        # if the last trial was a correction trial and we didn't get it correct,
        # then this is a correction trial too
        if (self.correction_trial or self.last_was_correction) and not self.correct:
            return True
        # if the last trial was a correction trial and we just corrected, no correction trial this time
        elif (self.correction_trial or self.last_was_correction) and self.correct:
            self.last_was_correction = False
            return False
        # if last trial was not a correction trial we spin  *to test* for one
        elif np.random.rand() < self.correction_pct:
            self.last_was_correction = True
            return False
        else:
            return False

    def update(self, response, correct):
        """
        At the end of a trial, update the status of our internal variables
        with the outcome of the trial.

        Args:
            response ('L', 'R'): How the subject responded
            correct (0, 1): Whether the response was correct.
        """
        self.response = response
        self.correct = correct
        if self.bias:
            self.bias.update(response, self.target)

    def end(self):
        """
        End all of our stim. Stim should have an `.end()` method of their own

        """

        for side, v in self.stimuli.items():
            for stim in v:
                try:
                    stim.end()
                except AttributeError:
                    print('stim does not have an end method! \n{}'.format(str(stim)))

class Proportional(Stim_Manager):
    """
    Present groups of stimuli with a particular frequency.

    Frequencies do not need to add up to 1, groups will be selected with the frequency
    (frequency)/(sum(frequencies)).

    Arguments:
        stim (dict): Dictionary with the structure:
            ::

                {'manager': 'proportional',
                 'type': 'sounds',
                 'groups': (
                     {'name':'group_name',
                      'frequency': 0.2,
                      'sounds':{
                          'L': [{Tone1_params}, {Tone2_params}...],
                          'R': [{Tone3_params}, {Tone4_params}...]
                      }
                    },
                    {'name':'second_group',
                      'frequency': 0.8,
                      'sounds':{
                          'L': [{Tone1_params}, {Tone2_params}...],
                          'R': [{Tone3_params}, {Tone4_params}...]
                      }
                    })
                }

    Attributes:
        stimuli (dict): A dictionary of stimuli organized into groups
        groups (dict): A dictionary mapping group names to frequencies

    """

    def __init__(self, stim):
        super(Proportional, self).__init__()

        self.stimuli = {}

        self.frequency_type = None

        if stim['type'] == 'sounds':
            if 'groups' in stim.keys():
                self.frequency_type = "within_groups"
                # top-level groups, choose group then choose side
                self.init_sounds_grouped(stim['groups'])
                self.store_groups(stim)
            else:
                self.frequency_type = "within_side"
                # second-level frequencies, side is chosen and then
                # probability from within a side
                self.init_sounds_individual(stim['sounds'])



    def init_sounds_grouped(self, sound_stim):
        """
        Instantiate sound objects similarly to :class:`.Stim_Manager`, just organizes them into groups.

        Args:
            sound_stim (tuple, list): an iterator like:
                ::

                    (
                     {'name':'group_name',
                      'frequency': 0.2,
                      'sounds': {
                          'L': [{Tone1_params}, {Tone2_params}...],
                          'R': [{Tone3_params}, {Tone4_params}...]
                        }
                    },
                    {'name':'second_group',
                      'frequency': 0.8,
                      'sounds':{
                          'L': [{Tone1_params}, {Tone2_params}...],
                          'R': [{Tone3_params}, {Tone4_params}...]
                      }
                    })
        """
        # Iterate through sounds and load them to memory
        self.stimuli = {}

        if isinstance(sound_stim, tuple) or isinstance(sound_stim, list):
            for group in sound_stim:
                group_name = group['name']

                # instantiate sounds
                self.stimuli[group_name] = {}
                for k, v in group['sounds'].items():
                    if isinstance(v, list):
                        self.stimuli[group_name][k] = []
                        for sound in v:
                            # We send the dict 'sound' to the function specified by 'type' and '
                            # ' as kwargs
                            self.stimuli[group_name][k].append(autopilot.get('sound', sound['type'])(**sound))
                    # If not a list, a single sound
                    else:
                        self.stimuli[group_name][k] = [autopilot.get('sound', v['type'])(**v)]


    def init_sounds_individual(self, sound_stim):
        """
        Initialize sounds with individually set presentation frequencies.

        .. todo::

            This method reflects the need for managers to have a unified schema,
            which will be built in a future release of Autopilot.

        Args:
            sound_stim (dict): Dictionary of {'side':[sound_params]} to generate sound stimuli

        Returns:

        """
        self.stim_freqs = {}
        for side, sound_params in sound_stim.items():
            self.stimuli[side] = []
            self.stim_freqs[side] = []
            if isinstance(sound_params, list):
                for sound in sound_params:
                    self.stimuli[side].append(autopilot.get('sound', sound['type'])(**sound))
                    self.stim_freqs[side].append(float(sound['management']['frequency']))
            else:
                self.stimuli[side].append(autopilot.get('sound', sound_params['type'])(**sound_params))
                self.stim_freqs[side].append(float(sound_params['management']['frequency']))

        # normalize frequencies within sides to sum to 1
        for side, freqs in self.stim_freqs.items():
            side_sum = np.sum(freqs)
            self.stim_freqs[side] = tuple([float(f)/side_sum for f in freqs])

    def store_groups(self, stim):
        """
        store groups and frequencies
        """
        self.groups = {}
        # also store (later) as tuples for choice b/c dicts are unordered
        group_names = []
        group_freqs = []
        for group in stim['groups']:
            group_name = group['name']
            frequency = float(group['frequency'])
            self.groups[group_name] = frequency
            group_names.append(group_name)
            group_freqs.append(frequency)

        self.group_names = tuple(group_names)
        group_freqs = np.array(group_freqs).astype(np.float)
        group_freqs = group_freqs/np.sum(group_freqs)

        self.group_freqs = tuple(group_freqs)



    def set_triggers(self, trig_fn):
        """
        Give a callback function to all of our stimuli for when the stimulus ends.

        Note:
            Stimuli need a `set_trigger` method.

        Args:
            trig_fn (callable): A function to be given to stimuli via `set_trigger`
        """
        # set a callback function for when the stimulus ends

        if self.frequency_type == "within_group":
            for _, group in self.stimuli.items():
                    for _, v in group.items():
                        for astim in v:
                            astim.set_trigger(trig_fn)
        elif self.frequency_type == "within_side":
            for side, sound_group in self.stimuli.items():
                for astim in sound_group:
                    astim.set_trigger(trig_fn)

        else:
            ValueError('Dont know how to set triggers')


    def next_stim(self):
        """
        Compute and return the next stimulus

        If we are doing correction trials, compute that.

        Same thing with bias correction.

        Otherwise, randomly select a stimulus to present, weighted by its group frequency.

        Returns:
            ('L'/'R' Target, 'L'/'R' distractor, Stimulus to present)
        """
        # compute and return the next stim

        # first: if we're doing correction trials, compute that
        if self.correction:
            self.correction_trial = self.compute_correction()
            if self.correction_trial:
                return self.target, self.distractor, self.last_stim

        # otherwise we check for bias correction
        # it will return a threshold for random choice
        if self.bias:
            threshold = self.bias.next_bias()
        else:
            threshold = 0.5

        # choose side
        if np.random.rand()<threshold:
            self.target = 'L'
        else:
            self.target = 'R'

        if self.target == 'L':
            self.distractor = 'R'
        elif self.target == 'R':
            self.distractor = 'L'

        if self.frequency_type == "within_group":

            # pick a stimulus based on group frequency
            group = np.random.choice(self.group_names, p=self.group_freqs)
            # within that group pick a random stimulus
            self.last_stim = np.random.choice(self.stimuli[group][self.target])

        elif self.frequency_type == "within_side":
            self.last_stim = np.random.choice(self.stimuli[self.target],
                                              p=self.stim_freqs[self.target])
        else:
            ValueError('Dont know what freq type we are')

        return self.target, self.distractor, self.last_stim





class Bias_Correction(object):
    """
    Basic Bias correction module. Modifies the threshold of random stimulus
    choice based on history of biased responses.

    Attributes:
        responses (:class:`collections.deque`): History of prior responses
        targets (:class:`collections.deque`): History of prior targets.
    """

    def __init__(self, mode='thresholded_linear', thresh=.2, window=100):
        """
        Args:
            mode: One of the following:

                * `'thresholded linear'` : above some threshold, do linear bias correction
                    eg. if response rate 65% left, make correct be right 65% of the time

            thresh (float): threshold above chance, ie. 0.2 means has to be 70% biased in window
            window (int): number of trials to calculate bias over
        """
        self.mode = mode
        self.threshold = float(thresh)
        self.window = int(window)
        self.responses = deque(maxlen=self.window)
        self.targets = deque(maxlen=self.window)

    def next_bias(self):
        """
        Compute the next bias depending on `self.mode`

        Returns:
            float: Some threshold :class:`.Stim_Manager` uses to decide left vs right.

        """

        # compute the bias threshold
        if self.mode == 'thresholded_linear':
            return self.thresholded_linear()

    def thresholded_linear(self):
        """
        If we are above the threshold, linearly correct the rate of
        presentation to favor the rarely responded side.

        eg. if response rate 65% left, make correct be right 65% of the time

        Returns:
            float: 0.5-bias, where bias is the difference between the mean response and mean target.
        """
        # if we are above threshold, return the bias
        bias = np.mean(self.responses)-np.mean(self.targets)
        if np.abs(bias)>self.threshold:
            return 0.5-bias
        else:
            return 0.5


    def update(self, response, target):
        """
        Store some new response and target values

        Args:
            response ('R', 'L'): Which side the subject responded to
            target ('R', 'L'): The correct side.
        """
        if isinstance(response, str):
            if response == "R":
                response = 1.0
            elif response == "L":
                response = 0.0

        if isinstance(target, str):
            if target == "R":
                target = 1.0
            elif target == "L":
                target = 0.0

        self.responses.append(float(response))
        self.targets.append(float(target))

MANAGER_MAP = {
    'proportional': Proportional
}





# class Reward_Manager(object):
#     def __init__(self):
#         pass



import platform
import os
import codecs

#from skbuild import setup, constants
from setuptools import find_packages, setup
import subprocess
import sys

# fix user install issue
import site
site.ENABLE_USER_SITE = "--user" in sys.argv[1:]

# declare defaults
IS_RASPI = False
SCRIPTS = []
PACKAGES = []
# CMAKE_ARGS = ['-DCMAKE_BUILD_DIR={}'.format(constants.CMAKE_BUILD_DIR()),
#               '-DCMAKE_INSTALL_DIR={}'.format(constants.CMAKE_INSTALL_DIR()),
#               '-DSETUPTOOLS_INSTALL_DIR={}'.format(constants.SETUPTOOLS_INSTALL_DIR()),
#               '-DSKBUILD_DIR={}'.format(constants.SKBUILD_DIR()),
#               '-DCMAKE_BUILD_RPATH_USE_ORIGIN=1', # use relative paths in Rpaths
#               '-DCMAKE_INSTALL_RPATH=$ORIGIN/lib;$ORIGIN/../lib'
#               ]

REQUIREMENTS = []



# detect if on raspberry pi
try:
    ret = subprocess.call(['grep', '-q', 'BCM', '/proc/cpuinfo'])
    if ret == 0:
        IS_RASPI = True
except:
    pass


# detect architecture
_ARCH = platform.uname().machine
ARCH = None
if _ARCH in ('armv7l',):
    ARCH = "ARM32"
elif _ARCH in ('aarch64',):
    ARCH = 'ARM64'
elif _ARCH in ('x86_64',):
    ARCH = "x86"


def load_requirements(req_file):
    with open(req_file, 'r') as f:
        requirements = f.read().splitlines()

    # filter commented reqs
    requirements = [r for r in requirements if not r.startswith('#')]

    # filter empty reqs
    requirements = [r for r in requirements if r]

    return requirements

# configure for raspberry pi
if IS_RASPI:
    REQUIREMENTS = load_requirements('requirements/requirements_pilot.txt')

elif ARCH == 'x86':
    # is a terminal,
    REQUIREMENTS = load_requirements('requirements/requirements_terminal.txt')

elif os.environ.get('TRAVIS', False):
    REQUIREMENTS = load_requirements('requirements/requirements_tests.txt')

else:
    REQUIREMENTS = load_requirements('requirements.txt')

# add external packages that wouldn't get detected normally
packs = find_packages()
packs.extend(PACKAGES)

# load readme
with open(os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'README.md'), 'r') as readme_f:
    readme = readme_f.read()


# get package version
def read(rel_path):
    """
    https://packaging.python.org/guides/single-sourcing-package-version/
    """
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    """
    https://packaging.python.org/guides/single-sourcing-package-version/
    """
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

setup(
    name="auto-pi-lot",
    version=get_version('autopilot/__init__.py'),
    description="Distributed behavioral experiments",
    long_description = readme,
    long_description_content_type='text/markdown',
    author="Jonny Saunders",
    author_email="j@nny.fyi",
    url="https://auto-pi-lot.com",
    license="MPL-2.0",
    scripts = SCRIPTS,
    packages=packs,
    #cmake_args=CMAKE_ARGS,
    install_requires = REQUIREMENTS,
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Topic :: Scientific/Engineering"
    ],
    python_requires=">=3.7.*"
)



import base64
import datetime
import json

import blosc


class Message(object):
    """
    A formatted message that takes ``value``, sends it to ``id``, who should call
    the listen method indicated by the ``key``.

    Additional message behavior can be indicated by passing ``flags``

    Numpy arrays given in the value field are automatically serialized and deserialized
    when sending and receiving using bas64 encoding and blosc compression.

    `id`, `to`, `sender`, and `key` are required attributes,
    but any other key-value pair passed on init is added to the message's attributes
    and included in the message.

    Can be indexed and set like a dictionary (message['key'], etc.)


    Attributes:
        id (str): ID that uniquely identifies a message.
            format {sender.id}_{number}
        to (str): ID of socket this message is addressed to
        sender (str): ID of socket where this message originates
        key (str): Type of message, used to select a listen method to process it
        value: Body of message, can be any type but must be JSON serializable.
        timestamp (str): Timestamp of message creation
        ttl (int): Time-To-Live, each message is sent this many times at max,
            each send decrements ttl.
        flags (dict): Flags determine additional message behavior. If a flag has no value associated with it,
            add it as a key with ``None`` as the value (eg. self.flags['MINPRINT'] = None), the value doesn't matter.

            * ``MINPRINT`` - don't print the value in logs (eg. when a large array is being sent)
            * ``NOREPEAT`` - sender will not seek, and recipients will not attempt to send message receipt confirmations
            * ``NOLOG`` - don't log this message! for streaming, or other instances where the constant printing of the logger is performance prohibitive
    """

    def __init__(self, msg=None, expand_arrays = False,  **kwargs):
        # Messages don't need to have all attributes on creation,
        # but do need them to serialize
        """
        Args:
            *args:
            **kwargs:
        """
        self.id = None  # number of message, format {sender.id}_{number}
        self.to = None
        self.sender = None
        self.key = None
        # value is the only attribute that can be left None,
        # ie. with signal-type messages like "STOP"
        self.value = None
        self.timestamp = None
        self.flags = {}
        self.changed = False
        self.serialized = None

        # optional attrs should be instance attributes so they are caught by _-dict__
        self.flags = {}
        self.timestamp = None

        self.ttl = kwargs.get('ttl', 2)

        #set_trace(term_size=(120,40))
        #if len(args)>1:
        #    Exception("Messages can only be constructed with a single positional argument, which is assumed to be a serialized message")
        #elif len(args)>0:
        if msg:
            self.serialized = msg
            if expand_arrays:
                deserialized = json.loads(msg, object_pairs_hook=self._deserialize_numpy)
            else:
                deserialized = json.loads(msg)
            kwargs.update(deserialized)

        for k, v in kwargs.items():
            setattr(self, k, v)
            #self[k] = v

        # if we're not a previous message being recreated, get a timestamp for our creation
        if 'timestamp' not in kwargs.keys():
            self.get_timestamp()

        # self.DETECTED_MINPRINT = False

    def __str__(self):
        # type: () -> str
        # if len(str(self.value))>100:
        #     self.DETECTED_MINPRINT = True
        # TODO: Make verbose/debugging mode, print value in that case.
        if self.key == 'FILE' or ('MINPRINT' in self.flags.keys()):
            me_string = "ID: {}; TO: {}; SENDER: {}; KEY: {}, FLAGS: {}".format(self.id, self.to, self.sender, self.key, self.flags)
        else:
            me_string = "ID: {}; TO: {}; SENDER: {}; KEY: {}; FLAGS: {}; VALUE: {}".format(self.id, self.to, self.sender, self.key, self.flags, self.value)
        #me_string = "ID: {}; TO: {}; SENDER: {}; KEY: {}".format(self.id, self.to, self.sender, self.key)

        return me_string

    # enable dictionary-like behavior
    def __getitem__(self, key):
        """
        Args:
            key:
        """
        #value = self._check_dec(self.__dict__[key])
        # TODO: Recursively walk looking for 'NUMPY ARRAY' and expand before giving
        return self.__dict__[key]

    def __setitem__(self, key, value):
        """
        Args:
            key:
            value:
        """
        # self.changed=True
        #value = self._check_enc(value)
        self.__dict__[key] = value

    # def __setattr__(self, key, value):
    #     self.changed=True
    #     #value = self._check_enc(value)
    #     super(Message, self).__setattr__(self, key, value)
    #     self.__dict__[key] = value

    # def __getattr__(self, key):
    #     #value = self._check_dec(self.__dict__[key])
    #     return self.__dict__[key]
    #
    # def _check_enc(self, value):
    #     if isinstance(value, np.ndarray):
    #         value = json_tricks.dumps(value)
    #     elif isinstance(value, dict):
    #         for k, v in value.items():
    #             value[k] = self._check_enc(v)
    #     elif isinstance(value, list):
    #         value = [self._check_enc(v) for v in value]
    #     return value
    #
    # def _check_dec(self, value):
    #
    #     # if numpy array, reconstitute
    #     if isinstance(value, basestring):
    #         if value.startswith('{"__ndarray__'):
    #             value = json_tricks.loads(value)
    #     elif isinstance(value, dict):
    #         for k, v in value.items():
    #             value[k] = self._check_dec(v)
    #     elif isinstance(value, list):
    #         value = [self._check_dec(v) for v in value]
    #     return value

    def _serialize_numpy(self, array):
        """
        Serialize a numpy array for sending over the wire

        Args:
            array:

        Returns:

        """
        compressed = base64.b64encode(blosc.pack_array(array)).decode('ascii')
        return {'NUMPY_ARRAY': compressed}


    def _deserialize_numpy(self, obj_pairs):
        # print(len(obj_pairs), obj_pairs)
        if (len(obj_pairs) == 1) and obj_pairs[0][0] == "NUMPY_ARRAY":
            return blosc.unpack_array(base64.b64decode(obj_pairs[0][1]))
        else:
            return dict(obj_pairs)

    def expand(self):
        """
        Don't decompress numpy arrays by default for faster IO, explicitly expand them when needed

        :return:
        """
        pass






    def __delitem__(self, key):
        """
        Args:
            key:
        """
        self.changed=True
        del self.__dict__[key]

    def __contains__(self, key):
        """
        Args:
            key:
        """
        return key in self.__dict__

    def __len__(self):
        return len(self.__dict__)

    def get_timestamp(self):
        """
        Get a Python timestamp

        Returns:
            str: Isoformatted timestamp from ``datetime``
        """
        self.timestamp = datetime.datetime.now().isoformat()

    def validate(self):
        """
        Checks if `id`, `to`, `sender`, and `key` are all defined.

        Returns:
            bool (True): Does message have all required attributes set?
        """
        valid = True
        for prop in (self.id, self.to, self.sender, self.key):
            if prop is None:
                valid = False
        return valid




    def serialize(self):
        """
        Serializes all attributes in `__dict__` using json.

        Returns:
            str: JSON serialized message.
        """

        if not self.changed and self.serialized:
            return self.serialized

        valid = self.validate()
        if not valid:
            Exception("""Message invalid at the time of serialization!\n {}""".format(str(self)))
            return False

        # msg = {
        #     'id': self.id,
        #     'to': self.to,
        #     'sender': self.sender,
        #     'key': self.key,
        #     'value': self.value
        # }
        msg = self.__dict__
        # exclude 'serialized' so it's not in there twice
        try:
            del msg['serialized']
        except KeyError:
            pass

        try:
            msg_enc = json.dumps(msg, default=self._serialize_numpy).encode('utf-8')
            self.serialized = msg_enc
            self.changed=False
            return msg_enc
        except:
            return False


"""
Hardware that uses the GPIO pins of the Raspi. These classes rely on `pigpio
<http://abyz.me.uk/rpi/pigpio/>`_, whose daemon (pigpiod) must be running in the background --
typically this is handled with a launch script/system daemon (see the launch_pilot.sh script generated by setup_autopilot.py)

Autopilot uses a custom version of pigpio (`<https://github.com/sneakers-the-rat/pigpio>`_) that
returns isoformatted timestamps rather than tick numbers in callbacks. See the ``setup_pilot.sh`` script.

Note:
    Autopilot uses the "Board" rather than "Broadcom" numbering system, see :ref:`the numbering note. <numbering-note>`
    :class:`.GPIO` objects convert internally between board and bcm numbers using :attr:`.GPIO.pin` ,
    :attr:`.GPIO.pin_bcm` , :data:`.BOARD_TO_BCM` , and :data:`.BCM_TO_BOARD` .


Note:
    This module does not include hardware that uses the GPIO pins over a specific protocol like i2c
"""
import os
import sys
import threading
import time
import numpy as np
from datetime import datetime
import itertools
import typing
import warnings


from autopilot import prefs
from autopilot.hardware import Hardware, BOARD_TO_BCM
from autopilot import external

ENABLED = False
"""
False if pigpio cannot be imported -- and GPIO devices cannot be used.

True if pigpio can be imported
"""
try:
    import pigpio


    TRIGGER_MAP = {
        'U': pigpio.RISING_EDGE,
        1: pigpio.RISING_EDGE,
        True: pigpio.RISING_EDGE,
        'D': pigpio.FALLING_EDGE,
        0: pigpio.FALLING_EDGE,
        False: pigpio.FALLING_EDGE,
        'B': pigpio.EITHER_EDGE,
        (0,1): pigpio.EITHER_EDGE
    }
    """
    Maps user input descriptions of triggers to the corresponding pigpio object.
    """

    INVERSE_TRIGGER_MAP = {
        pigpio.RISING_EDGE: 'U',
        pigpio.FALLING_EDGE: 'D',
        pigpio.EITHER_EDGE: 'B'
    }
    """
    Inverse of :data:`.TRIGGER_MAP`. Used to assign canonical references to triggers -- 
    ie. it is possible to take multiple params (1, True, 'U') -> pigpio trigger objects,
    but there is one preferred way to refer to a pigpio object.
    """

    PULL_MAP = {
        1: pigpio.PUD_UP,
        True: pigpio.PUD_UP,
        'U': pigpio.PUD_UP,
        0: pigpio.PUD_DOWN,
        False: pigpio.PUD_DOWN,
        'D': pigpio.PUD_DOWN,
        None: pigpio.PUD_OFF
    }
    """
    Maps user input descriptions of internal resistor pullups/downs to the corresponding
    pigpio object.
    """

    INVERSE_PULL_MAP = {
        pigpio.PUD_UP: 'U',
        pigpio.PUD_DOWN: 'D',
        pigpio.PUD_OFF: None
    }
    """
    Inverse of :data:`.PULL_MAP`, mapping pigpio objects for internal resistor pullups/downs to 
    their canonical form ('U', 'D', None for pullup, pulldown, or no pull)
    """

    ENABLED = True

except ImportError:
    if prefs.get('AGENT') == "PILOT":
        warnings.warn("pigpio could not be imported, gpio not enabled", ImportWarning)


def clear_scripts(max_scripts=256):
    """
    Stop and delete all scripts running on the pigpio client.

    To be called, eg. between tasks to ensure none are left hanging by badly behaved GPIO devices

    Args:
        max_scripts (int): maximum number of scripts allowed by pigpio. Set in ``pigpio.c`` and not exported
            to the python module, so have to hardcode it again here, default for pigpio fork is 256
    """

    if globals()['ENABLED'] == False:
        warnings.warn('pigpio was not imported, so scripts cannot be cleared')
        return

    pig = pigpio.pi()

    for i in range(max_scripts):
        try:
            pig.stop_script(i)
        except pigpio.error as e:
            if 'unknown script id' in str(e):
                continue

        try:
            pig.delete_script(i)
        except Exception as e:
            warnings.warn(f'Error deleting script {i}, got exception: \n{e}')








class GPIO(Hardware):
    """
    Metaclass for hardware that uses GPIO. Should not be instantiated on its own.

    Handles initializing pigpio and wraps some of its commonly used methods

    Args:
        pin (int): The `Board-numbered <https://raspberrypi.stackexchange.com/a/12967>`_ GPIO pin of this object.
        polarity (int): Logic direction. if 1: on=High=1, off=Low=0; if 0: off=Low=0, on=High=1
        pull (str, int): state of pullup/down resistor. Can be set as 'U'/'D' or 1/0 to pull up/down. See :data:`.PULL_MAP`
        trigger (str, int, bool): whether callbacks are triggered on rising ('U', 1, True), falling ('D', 0, False),
            or both edges ('B', (0,1))
        kwargs: passed to the :class:`.Hardware` superclass.

    Attributes:
        pig (:class:`pigpio.pi`): An object that manages connection to the pigpio daemon. See docs at http://abyz.me.uk/rpi/pigpio/python.html
        CONNECTED (bool): Whether the connection to pigpio was successful
        pigpiod: Reference to the pigpiod process launched by :func:`.external.start_pigpiod`
        pin (int): The `Board-numbered <https://raspberrypi.stackexchange.com/a/12967>`_ GPIO pin of this object.
        pin_bcm (int): The BCM number of the connected pin -- used by pigpio. Converted from pin passed as argument on initialization,
            which is assumed to be the board number.
        pull (str, int): state of pullup/down resistor. Can be set as 'U'/'D' or 1/0 to pull up/down
        polarity (int): Logic direction. if 1: on=High=1, off=Low=0; if 0: off=Low=0, on=High=1
        on (int): if polarity == 1, high/1. if polarity == 0, low/0
        off (int): if polarity == 1, low/0. if polarity == 0, high/1
        trigger (str, int, bool): whether callbacks are triggered on rising ('U', 1, True), falling ('D', 0, False),
            or both edges ('B', (0,1))
        trigger_edge: The pigpio object representing RISING_EDGE, FALLING_EDGE, BOTH_EDGES. Set by :attr`.trigger`

    """

    def __init__(self, pin=None, polarity=1, pull = None, trigger = None, **kwargs):
        super(GPIO, self).__init__(**kwargs)

        if not ENABLED:
            raise RuntimeError('Couldnt import pigpio, so GPIO objects cant be used')

        # initialize attributes
        self._polarity = None
        self._pull = None
        self._pin = None
        self._trigger = None
        self.trigger_edge = None
        try:
            self.pin_bcm = None
        except AttributeError:
            # if a subclass has made this a property, don't fail here
            self.logger.warning('pin_bcm is defined as a property without a setter so cant be set')

        self.pig = None # type: typing.Optional[pigpio.pi]
        self.pigpiod = None

        # init pigpio
        self.CONNECTED = False
        self.CONNECTED = self.init_pigpio()

        # set default attributes

        self.pin = pin
        self.polarity = polarity
        self.pull = pull
        self.trigger = trigger

        if not self.CONNECTED:
            RuntimeError('No connection could be made to the pigpio daemon')

    def init_pigpio(self) -> bool:
        """
        Create a socket connection to the pigpio daemon and set as :attr:`GPIO.pig`

        Returns:
            bool: True if connection was successful, False otherwise
        """
        self.pigpiod = external.start_pigpiod()
        self.pig = pigpio.pi()
        if self.pig.connected:
            return True
        else:
            return False

    @property
    def pin(self):
        """
        `Board-numbered <https://raspberrypi.stackexchange.com/a/12967>`_ GPIO pin.

        When assigned, also updates :attr:`.pin_bcm` with the BCM-numbered pin.
        """
        return self._pin

    @pin.setter
    def pin(self, pin):
        if pin is None:
            return
        assert(int(pin))

        self._pin = int(pin)
        self.pin_bcm = BOARD_TO_BCM[self._pin]

    @property
    def state(self) -> bool:
        """
        Instantaneous state of GPIO pin, on (``True``) or off (``False``)

        Returns:
            bool
        """
        return bool(self.pig.read(self.pin_bcm))

    @property
    def pull(self):
        """
        State of internal pullup/down resistor.

        See :data:`.PULL_MAP` for possible values.

        Returns:
            int: 'U'/'D'/None for pulled up, down or not set.
        """
        return self._pull

    @pull.setter
    def pull(self, direction):

        try:
            pull = PULL_MAP[direction]
            self._pull = INVERSE_PULL_MAP[pull]
            self._pull = direction
            self.pig.set_pull_up_down(self.pin_bcm, pull)
        except KeyError:
            ValueError("pull must be one of 0/'D'/False for pulldown, 1/'U'/True for pullup, or None to clear")

    @property
    def polarity(self):
        """
        Logic direction. if 1: on=High=1, off=Low=0; if 0: off=Low=0, on=High=1.

        When set, updates :attr:`~.GPIO.on` and :attr:`~.GPIO.off` accordingly
        """
        return self._polarity

    @polarity.setter
    def polarity(self, polarity):
        if polarity == 1:
            self.on = 1
            self.off = 0
        elif polarity == 0:
            self.on = 0
            self.off = 1
        else:
            raise ValueError('polarity must be 0 or 1')

        self._polarity = polarity

    @property
    def trigger(self):
        """
        dict: Maps strings (('U',1,True), ('D',0,False), ('B',[0,1])) to pigpio edge types
        (RISING_EDGE, FALLING_EDGE, EITHER_EDGE), respectively.
        """
        return self._trigger


    @trigger.setter
    def trigger(self, trigger):
        try:
            self.trigger_edge = TRIGGER_MAP[trigger]
            self._trigger = INVERSE_TRIGGER_MAP[self.trigger_edge]
        except KeyError:
            ValueError('trigger must be one of {}. instead got {}'.format(INVERSE_TRIGGER_MAP.values(),
                                                                          trigger))

    def release(self):
        """
        Release the connection to the pigpio daemon.

        Note:
            the Hardware metaclass will call this method on object deletion.
        """
        try:
            self.pull = None
        except:
            pass

        self.pig.stop()


class Digital_Out(GPIO):
    """
    TTL/Digital logic out through a GPIO pin.

    Args:
        pin (int): The `Board-numbered <https://raspberrypi.stackexchange.com/a/12967>`_ GPIO pin of this object
        pulse_width (int): Width of digital output :meth:`~.Digital_Out.pulse` (us). range: 1-100
        polarity (bool): Whether 'on' is High (1, default) and pulses bring the voltage High, or vice versa (0)

    Attributes:
        scripts (dict): maps script IDs to pigpio script handles
        pigs_function (bytes): when using pigpio scripts, what function is used to set the value of the output?
            (eg. 'w' for digital out, 'gdc' for pwm, more info here: http://abyz.me.uk/rpi/pigpio/pigs.html)
        script_counter (:class:`itertools.count`): generate script IDs if not explicitly given to :meth:`~.Digital_Out.series`.
            generated IDs are of the form 'series_#'
    """

    output = True
    type="DIGITAL_OUT"
    pigs_function = b"w"

    def __init__(self, pin=None, pulse_width=100, polarity=1, **kwargs):
        """

        """
        super(Digital_Out, self).__init__(pin, polarity=polarity, **kwargs)

        self._last_script = None
        self.scripts = {}
        self.script_handles = {}
        self.script_counter = itertools.count()

        self.pulse_width = np.clip(pulse_width, 0, 100).astype(np.int)
        if pulse_width > 100 or pulse_width < 0:
            Warning('pulse_width must be <100(ms) & >0 and has been clipped to {}'.format(self.pulse_width))

        # setup pin if we are a true single-pin digital output and not a subclass
        if isinstance(self.pin_bcm, int):

            self.pig.set_mode(self.pin_bcm, pigpio.OUTPUT)
            self.set(self.off)

    def set(self, value: bool):
        """
        Set pin logic level.

        Default uses :meth:`pigpio.pi.write`, but can be overwritten by inheriting classes

        Stops the last running script when called.

        Args:
            value (int, bool): (1, True) to set High, (0, False) to set Low.
        """
        self.stop_script()

        self.pig.write(self.pin_bcm, value)

    def turn(self, direction='on'):
        """
        Change output state using on/off parlance. logic direction varies based on :attr:`Digital_Out.polarity`

        Stops the last running script when called.

        Args:
            direction (str, bool): 'on', 1, or True to turn to :attr:`~.Digital_Out.on` and vice versa for :attr:`~.Digital_Out.off`
        """

        self.stop_script()

        if direction in ('on', True, 1):
            self.set(self.on)
        elif direction in ('off', False, 0):
            self.set(self.off)

    def toggle(self):
        """
        If pin is High, set Low, and vice versa.

        Stops the last running script when called.
        """

        self.stop_script()

        if self.pig.read(self.pin_bcm):
            self.set(0)
        else:
            self.set(1)


    def pulse(self, duration=None):
        """
        Send a timed :attr:`~Digital_Out.on` pulse.

        Args:
            duration (int): If None (default), uses :attr:`~.Digital_Out.duration`, otherwise
                duration of pulse from 1-100us.
        """

        if not duration:
            self.pig.gpio_trigger(self.pin_bcm, self.pulse_width, self.on)

        elif duration:
            duration = np.clip(duration, 0, 100).astype(np.int)
            self.pig.gpio_trigger(self.pin_bcm,
                                  duration,
                                  self.on)

        else:
            raise RuntimeError('How did we even get here? Pulse value must be an int 1-100')


    def _series_script(self, values, durations = None, unit="ms", repeat=None, finish_off = True):
        """
        Create a pigpio script to set a pin to a series of values for a series of durations.

        Typically shouldn't be called by itself, is used by :meth:`~.Digital_Out.series` or  :meth:`~.Digital_Out.store_series`

        For more information on pigpio scripts, see: `<http://abyz.me.uk/rpi/pigpio/pigs.html#Scripts>`_

        Args:
            values (list): A list of tuples of (value, duration) or a list of values in (1,0) to set self.pin_bcm to.
            durations (list): If ``values`` is not a list of tuples, a list of durations.
                len(durations) must be either == len(values) or else len(durations) == 1,
                in which case the duration is repeated.
            unit ("ms", "us"): units of durations in milliseconds or microseconds
            repeat (int): If the script should be repeated, how many times? A value of 2 results in
                the script being run 2 times total, not 2 *additional* times (or, 3 total times)
            finish_off (bool): If true, the script ends by turning the pin to :attr:`~.Digital_Out.off`

        Returns:
            (str): the constructed script string
        """

        # if single value as an int/float/etc, wrap in list.
        if not (isinstance(values, list) or isinstance(values, tuple)):
            values = [values]


        if durations:
            if isinstance(durations, float) or isinstance(durations, int):
                durations = [round(durations)]

            if len(values) == len(durations):
                iter_series = zip(values, durations)
            elif len(durations) == 1:
                iter_series = itertools.product(values, durations)
            else:
                raise ValueError("length of  values and durations must be equal, or length of durations must be 1. got len(values)={}, len(durations)={}".format(len(values), len(durations)))
        else:
            iter_series = values.__iter__()


        if unit == "ms":
            wait_fn = "mils"
        elif unit == "us":
            wait_fn = "mics"
        else:
            raise ValueError("Unit for durations must be ms (milliseconds) or us (microseconds)")

        string_pieces = [b" ".join((self.pigs_function,
                                    str(self.pin_bcm).encode('utf-8'),
                                    str(val).encode('utf-8'),
                                    bytes(wait_fn, 'utf-8'),
                                    str(round(dur)).encode('utf-8'))) for val, dur in iter_series]
        script_str = b" ".join(string_pieces)

        if repeat:
            try:
                repeat = int(repeat)
            except:
                raise ValueError('Repeat must be coerceable to an integer, got {}'.format(repeat))

            script_str = b" ".join(("LD v0", str(repeat-1), # "LD (load) variable 0 with number of repeats"
                                    "tag 999",            # create a tag that can be returned to
                                    script_str,           # do the script one time
                                    "dcr v0"              # decrement v0. 
                                    "jp 999")             # jump to the tag at the beginning if v0 is still positive
                                   )

        # turn off when finished
        if finish_off:
            script_str += b" ".join((self.pigs_function, str(self.pin_bcm).encode('utf-8'), str(self.off).encode('utf-8')))


        return script_str

    def store_series(self, id, **kwargs):
        """
        Create, and store a pigpio script for a series of output values to be called by :meth:`~.Digital_Out.series`

        Args:
            id (str): shorthand key used to call this series with :meth:`.series`
            kwargs: passed to :meth:`~.Digital_Out._series_script`

        """

        # actually might want to overwrite script_handles w/ same ids, like a solenoid that changes its 'open' duration
        # if id in self.script_handles.keys():
        #     Exception("Script with id {} already created!".format(id))

        series_script = self._series_script(**kwargs)

        # check if there is an identical script already and use that instead
        matches = [script_id for script_id, test_script in self.scripts.items() if test_script == series_script]
        if len(matches)>0:
            script_id = self.script_handles[matches[0]]
        else:
            try:
                script_id = self.pig.store_script(series_script)
            except Exception as e:
                if 'illegal script command' in str(e):
                    raise Exception(f'got pigpio exception: {e} from attempted script {series_script}')
                else:
                    raise e

        self.script_handles[id] = script_id
        self.scripts[id] = series_script

    def series(self, id=None, delete=None, **kwargs):
        """
        Execute a script that sets the pin to a series of values for a series of durations.

        See :meth:`~.Digital_Out._series_script` for series parameterization.

        Ideally one would use :meth:`.store_series` and use the returned id to call this function.
        Otherwise, this method calls :meth:`.store_series` and runs it.

        Args:
            id (str, int): ID of the script, if not already created, created with :meth:`.store_script`.
                If None (default), an ID is generated with :attr:`~.Digital_Out.script_counter` of the form
                ``'script_#'``
            kwargs: passed to :meth:`~.Digital_Out._series_script`


        """

        return_id = False

        if id:
            if id not in self.script_handles.keys():
                self.store_series(id, **kwargs)

        # if we weren't called with an ID, have to have a list of values/etc to create a script
        else:
            id = "script_{}".format(next(self.script_counter))
            self.store_series(id, **kwargs)
            # if we haven't been explicitly told to not delete it, we should
            if delete is None:
                delete = True


            # since we've generated a script ID, we should return it
            return_id = True

        script_status = self.pig.script_status(self.script_handles[id])
        if script_status == pigpio.PI_SCRIPT_INITING:
            check_times = 0
            while self.pig.script_status(self.script_handles[id]) == pigpio.PI_SCRIPT_INITING:
                # TODO: Expose this as a parameter -- how long to try and init scripts before skipping, mebs some general 'timeout' variable for all blocking ops.
                time.sleep(0.005)
                check_times += 1
                if check_times > 200:
                    break

        try:
            self.pig.run_script(self.script_handles[id])
            self._last_script = id
        except pigpio.error as e:
            self.logger.exception(f'Couldnt run script: {e}')
        finally:
            if delete:
                self.delete_script(id)


        if return_id:
            return id

    def delete_script(self, script_id):
        """
        spawn a thread to delete a script with id ``script_id``

        This is a 'soft' deletion -- it checks if the script is running, and waits for up to 10 seconds
        before actually deleting it.

        The script is deleted from the pigpio daemon, from ``script_handles`` and from ``scripts``

        Args:
            script_id (str): a script ID in :attr:`.Digital_Out.script_handles`
        """
        delete_script = threading.Thread(target=self._delete_script, args=(script_id,))
        delete_script.start()

    def _delete_script(self, script_id):
        checktimes = 0
        while self.pig.script_status(self.script_handles[script_id]) == pigpio.PI_SCRIPT_RUNNING:
            time.sleep(1)
            checktimes += 1
            if checktimes > 10:
                continue


        self.pig.delete_script(self.script_handles[script_id])

        del self.scripts[self.script_handles[script_id]]
        del self.script_handles[script_id]


    def delete_all_scripts(self):
        """
        Stop and delete all scripts

        This is a "hard" deletion -- the script will be immediately stopped if it's running.
        """

        for script_handle, script_id in self.script_handles.items():
            try:
                self.stop_script(script_handle)
            except Exception as e:
                self.logger.exception(e)

            try:
                self.pig.delete_script(script_id)
            except AttributeError:
                pass
            except Exception as e:
                if 'unknown script id' in str(e):
                    pass
                else:
                    self.logger.exception(e)

            try:
                del self.scripts[self.script_handles[script_id]]
            except KeyError:
                pass

            try:
                del self.script_handles[script_id]
            except KeyError:
                pass


    def stop_script(self, id=None):
        """
        Stops a running pigpio script

        Args:
            id (str, none): If None, stops the last run script. if str, stops script with that id.

        """

        if not id:
            id = self._last_script

        if id is None:
            # if there was no passed id and _last_script is None, return peacefully
            return

        # if we were passed a keyed, named script id, convert it to the pigio integer script id
        if id in self.script_handles.keys():
            script_id = self.script_handles[id]
            status, _ = self.pig.script_status(script_id)

            # if the script is still running, stop it and set the pin to off.
            if status == pigpio.PI_SCRIPT_RUNNING:
                self.pig.stop_script(script_id)
                #self.set(self.off)

            # set last_script to None so we won't try to double-stop this script
            self._last_script = None

        else:
            self.logger.warning('No script with id {} found to stop'.format(id))



    def release(self):
        """
        Stops and deletes all scripts, sets to :attr:`~.Digital_Out.off`, and calls :meth:`.GPIO.release`
        """
        try:

            self.delete_all_scripts()

            #self.stop_script()
            self.set(self.off)
            time.sleep(0.1)

        except AttributeError:
            # self.pig has already been deleted (release has already been called)
            # so self.pig has no attribute 'send' because it is None
            pass
        finally:
            super(Digital_Out, self).release()


class Digital_In(GPIO):
    """
    Record digital input and call one or more callbacks on logic transition.

    Args:
        pin (int): `Board-numbered <https://raspberrypi.stackexchange.com/a/12967>`_ GPIO pin.
        event (:class:`threading.Event`): For callbacks assigned with :meth:`.assign_cb` with ``evented = True``,
            set this event whenever the callback is triggered. Can be used to handle
            stage transition logic here instead of the :class:`.Task` object, as is typical.
        record (bool): Whether all logic transitions should be recorded as a list of ('EVENT', 'Timestamp') tuples.
        **kwargs: passed to :class:`GPIO`

    Sets the internal pullup/down resistor to :attr:`.Digital_In.off` and
    :attr:`.Digital_In.trigger` to :attr:`.Digital_In.on` upon instantiation.


    Note:
        pull and trigger are set by polarity on initialization in digital inputs, unlike other GPIO classes.
        They are not mutually synchronized however, ie. after initialization if any one of these attributes are changed, the other two will remain the same.


    Attributes:
        pig (:meth:`pigpio.pi`): The pigpio connection.
        pin (int): Broadcom-numbered pin, converted from the argument given on instantiation
        callbacks (list): A list of :meth:`pigpio.callback`s kept to clear them on exit
        polarity (int): Logic direction, if 1: off=0, on=1, pull=low, trigger=high and vice versa for 0
        events (list): if :attr:`.record` is True, a list of ('EVENT', 'TIMESTAMP') tuples
    """
    is_trigger=True
    type = 'DIGI_IN'
    input = True

    def __init__(self, pin, event=None, record=True, **kwargs):
        """

        """
        super(Digital_In, self).__init__(pin, **kwargs)

        # pull the resistor in the off direction and set the trigger to be the on direction
        self.pull = self.off
        self.trigger = self.on

        # We can be passed a threading.Event object if we want to handle stage logic here
        # rather than in the parent as is typical.
        self.event = event

        # List to store callback handles assigned with assign_cb
        self.callbacks = []

        # List to store logic transition events
        # FIXME: Should be a deque
        self.events = []

        self.record = record
        if self.record:
            self.assign_cb(self.record_event, add=True, evented=False, manual_trigger='B')

        # Setup pin
        self.pig.set_mode(self.pin_bcm, pigpio.INPUT)

    def assign_cb(self, callback_fn, add=True, evented=False, manual_trigger=None):
        """
        Sets ``callback_fn`` to be called when :attr:`.Digital_In.trigger` is detected.

        ``callback_fn`` must accept three parameters:

        * GPIO (int, 0-31): the BCM number of the pin that was triggered
        * level (0-2):

            * 0: change to low (falling)
            * 1: change to high (rising)
            * 2: no change (watchdog timeout)

        * timestamp (str): If using the Autopilot version of pigpio, an isoformatted timestamp

        Args:
            callback_fn (callable): The function to be called when triggered
            add (bool): Are we adding another callback?
                If False, the previous callbacks are cleared.
            evented (bool): Should triggering this event also set the internal :attr:`~.Digital_In.event`?
                Note that :attr:`.Digital_In.event` must have been passed.
            manual_trigger ('U', 'D', 'B'): Override :attr:`.Digital_In.trigger` if needed.
        """
        # If we aren't adding, we clear any existing callbacks
        if not add:
            self.clear_cb()
            if self.record:
                # if we're clearing all callbacks (maybe by accident)
                # but we're configured to record events, re-add the record cb.
                self.assign_cb(self.record_event, add=True, evented=False, manual_trigger="B")

        # We can set the direction of the trigger manually,
        # for example if we want to set 'BOTH' only sometimes
        if not manual_trigger:
            trigger_ud = self.trigger_edge
        else:
            trigger_ud = TRIGGER_MAP[manual_trigger]

        # We can handle eventing (blocking) here if we want (usually this is handled in the parent)
        # This won't work if we weren't init'd with an event.
        if evented and self.event:
            cb = self.pig.callback(self.pin_bcm, trigger_ud, self.event.set)
            self.callbacks.append(cb)
        elif evented and not self.event:
            raise Exception('We have no internal event to set!')

        cb = self.pig.callback(self.pin_bcm, trigger_ud, callback_fn)
        self.callbacks.append(cb)

    def clear_cb(self):
        """
        Tries to call `.cancel()` on each of the callbacks in :attr:`~Digital_In.callbacks`
        """
        for cb in self.callbacks:
            try:
                cb.cancel()
            except:
                self.logger.warning('could not cancel callback: {}'.format(cb))
        self.callbacks = []

    def record_event(self, pin, level, timestamp):
        """
        On either direction of logic transition, record the time

        Args:
            pin (int): BCM numbered pin passed from pigpio
            level (bool): High/Low status of current pin
            timestamp (str): isoformatted timestamp
        """

        # use self.pin rather than incoming pin b/c method is bound
        # (ie. only will be called by pin assigned to)
        # and self.pin is board rather than bcm numbered
        self.events.append((self.pin, level, timestamp))


    def release(self):
        """
        Clears any callbacks and calls :meth:`GPIO.release`
        """
        self.clear_cb()
        super(Digital_In, self).release()

class PWM(Digital_Out):
    """
    PWM output from GPIO.
    """
    output = True
    type="PWM"
    pigs_function = b'pwm'

    def __init__(self, pin, range=255, **kwargs):
        """
        Args:
            pin (int): Board numbered GPIO pin
            range (int): Maximum value of PWM duty-cycle. Default 255.
            **kwargs: passed to :class:`Digital_Out`
        """
        self._range = None
        super(PWM, self).__init__(pin, **kwargs)

        self.range = range

        #self.pig.set_mode(self.pin_bcm, pigpio.OUTPUT)
        # set to off in case of common anode/inverted polarity
        self.set(self.off)

    def set(self, value):
        """
        Sets PWM duty cycle normalized to :attr:`.polarity` and transformed by :meth:`._clean_value`

        Stops the last running script

        Args:
            value (int, float):
                - if int > 1, sets value (or :attr:`PWM.range`-value if :attr:`PWM.polarity` is inverted).
                - if 0 <= float <= 1, transforms to a proportion of :attr:`.range` (inverted  if needed as well).

        """
        value = self._clean_value(value)

        self.stop_script()

        self.pig.set_PWM_dutycycle(self.pin_bcm, value)

    def _clean_value(self, value):
        if value > 1:
            if value > self.range:
                self.logger.warning('clipping {} to range {}'.format(value, self.range))
                value = np.clip(value, 0, self.range).astype(np.int)

        elif 0 <= value <= 1:
            value = np.round(value * self.range).astype(np.int)

        else:
            self.logger.exception('PWM value must be an integer between 0 and range: {}, or a float between 0 and 1. got {}'.format(self.range, value))

        if self.polarity == 0:
            value = self.range - value

        return value

    @property
    def range(self):
        """
        Maximum value of PWM dutycycle.

        Doesn't set duration of PWM, but set values will be divided by this range.
        eg. if ``range == 200``, calling :meth:`PWM.set(100)` would result in a 50% duty cycle

        Args:
            (int): 25-40000
        """
        if not self._range:
            self._range = self.pig.get_PWM_range(self.pin_bcm)

        return self._range

    @range.setter
    def range(self, value):
        try:
            value = int(value)
        except ValueError:
            ValueError('PWM Range must be coerceable to an integer, got {}'.format(value))

        if (value < 25) or (value > 40000):
            Warning('PWM Range must be between 25 and 40000, got {}, clipping to range'.format(value))
            value = np.clip(value, 25, 40000).astype(np.int)

        self.pig.set_PWM_range(self.pin_bcm, value)
        self._range = value

        # set polarity again so off/on update
        self.polarity = self._polarity


    @property
    def polarity(self):
        """
        Logic direction.

        - if 1: on=High=:attr:`~PWM.range`, off=Low=0;
        - if 0: off=Low=0, on=High=:attr:`~PWM.range`.

        When set, updates :attr:`~PWM.on` and :attr:`~PWM.off`
        """
        return self._polarity

    @polarity.setter
    def polarity(self, polarity):
        if polarity == 1:
            self.on = self.range
            self.off = 0
        elif polarity == 0:
            self.on = 0
            self.off = self.range
        else:
            ValueError('polarity must be 0 or 1')
            return

        self._polarity = polarity

    def release(self):
        """
        Turn off and call :meth:`Digital_Out.release`

        Returns:

        """
        # FIXME: reimplementing parent release method here because of inconsistent use of self.off -- unify API and fix!!
        try:
            self.delete_all_scripts()
            self.set(0) # clean values should handle inversion, don't use self.off
            time.sleep(0.1)
            self.pig.stop()
        except AttributeError:
            # has already been called, pig already destroyed
            pass


class LED_RGB(Digital_Out):

    output = True
    type = "LEDS"

    def __init__(self, pins = None, r = None, g = None, b = None, polarity = 1, blink = True, **kwargs):
        """
        An RGB LED, wrapper around three :class:`PWM` objects.


        Args:
            pins (list): A list of (board) pin numbers.
                Either `pins` OR all `r`, `g`, `b` must be passed.
            r (int): Board number of Red pin - must be passed with `g` and `b`
            g (int): Board number of Green pin - must be passed with `r` and `b`
            b (int): Board number of Blue pin - must be passed with `r` and `g`:
            polarity (0, 1):
                0: common anode (low turns LED on)
                1: common cathode (low turns LED off)
            blink (bool): Flash RGB at the end of init to show we're alive and bc it's real cute.
            **kwargs: passed to :class:`Digital_Out`

        Attributes:
            channels (dict): The three PWM objects, {'r':PWM, ... etc}

        """
        # self._pin = {}
        self.channels = {}
        self.scripts = {}

        if 'pin' in kwargs.keys():
            raise ValueError('pin passed to LED_RGB, need a list of 3 pins instead (r, g, b). pin, used by single-pin GPIO objects, is otherwise ambiguous with pins.')

        super(LED_RGB, self).__init__(**kwargs)

        self.flash_params = None



        if pins and len(pins) == 3:
            for color, pin in zip(('r', 'g', 'b'), pins):
                self.channels[color] = PWM(pin, polarity=polarity, name=f"{self.name}_{color.upper()}")

        elif r and g and b:
            self.channels = {'r':PWM(r, polarity=polarity, name="{}_R".format(self.name)),
                             'g':PWM(g, polarity=polarity, name="{}_G".format(self.name)),
                             'b':PWM(b, polarity=polarity, name="{}_B".format(self.name))}
        else:
            raise ValueError('Must either set with pins= list/tuple of r,g,b pins, or pass all three as separate params')

        self.store_series(id='blink',
                          colors=((1,0,0),(0,1,0),(0,0,1),(0,0,0)),
                          durations=250)

        if blink:
            self.series(id='blink')

    @property
    def range(self) -> dict:
        """
        Returns:
            dict: ranges for each of the :attr:`LED_RGB.channels`
        """
        return {'r':self.channels['r'].range,
                'g':self.channels['g'].range,
                'b':self.channels['b'].range}

    @range.setter
    def range(self, range):
        for channel in self.channels.items():
            channel.range = range

    def set(self, value=None, r=None, g=None, b=None):
        """
        Set the color of the LED.

        Can either pass

        - a full (R, G, B) tuple to ``value``,
        - a single ``value`` that is applied to each channel,
        - if ``value`` is not passed, individual ``r``, ``g``, or ``b`` values can be passed (any combination can be set in a single call)

        Stops the last run script

        Args:
            value (int, float, tuple, list):  If list or tuple, an (R, G, B) color. If float or int, applied to each color channe.
                Can be set with floats 0-1, or ints >= 1  (See :attr:`PWM.range`).
                If None, use ``r``, ``g``, and ``b``.
            r (float, int): value to set red channel
            g (float, int): value to set green channel
            b (float, int): value to set blue channel
        """
        # Stop any running scripts (like blinks)
        self.stop_script()

        # Set either by `value` or by individual r, g, b
        if value is not None:
            # Set by `value`
            if isinstance(value, int) or isinstance(value, float):
                # Apply `value` to each channel
                for channel in self.channels.values():
                    channel.set(value)
            elif len(value) == 3:
                # Assume value is a tuple of r, g, b
                for channel_key, color_val in zip(('r','g','b'), value):
                    self.channels[channel_key].set(color_val)
            else:
                raise ValueError('Value must either be a single value or a tuple of (r,g,b)')
        else:
            # Set by individually specified r, g, b arguments
            if r is not None:
                self.channels['r'].set(r)
            if g is not None:
                self.channels['g'].set(g)
            if b is not None:
                self.channels['b'].set(b)

    def toggle(self):
        self.logger.warning('toggle not well defined for PWM')

    def pulse(self, duration=None):
        self.logger.warning('Use flash for LEDs instead')

    def _series_script(self, colors, durations = None, unit="ms", repeat=None, finish_off = True):
        """
        Create a script to flash a series of colors.

        Like :meth:`Digital_Out._series_script`, but sets all pins at once.

        Args:
            colors (list): a list of (R, G, B) colors, or a list of ((R,G,B),duration) tuples.
            durations (int, list): Duration of each color. if a single value, used for all ``colors``.
                if a list, ``len(durations) == len(colors)``. If None, colors must be ``((R,G,B),duration)`` tuples.
            unit ('ms', 'us'): unit of durations, milliseconds or microseconds
            repeat (int): Number of repetitions. If None, script runs once.
            finish_off (bool): Whether the channels should be set to ``off`` when the script completes

        Returns:
            str: constructed pigpio script string.
        """
        if durations:
            if isinstance(durations, int) or isinstance(durations, float):
                iter_series = itertools.product(colors, [round(durations)])
            elif len(colors) == len(durations):
                iter_series = zip(colors, durations)
            elif len(durations) == 1:
                iter_series = itertools.product(colors, durations)
            else:
                ValueError("length of  colors and durations must be equal, or length of durations must be 1. got len(colors)={}, len(durations)={}".format(len(colors), len(durations)))
        else:
            iter_series = colors.__iter__()

        if unit == "ms":
            wait_fn = "mils"
        elif unit == "us":
            wait_fn = "mics"
        else:
            ValueError("Unit for durations must be ms (milliseconds) or us (microseconds)")


        string_pieces = []
        for color, duration in iter_series:
            # iter over keys rather than values of channels dict b/c dicts are unordered and colors are an ordered tuple
            for channel_key, color_value in zip(('r', 'g', 'b'), color):
                string_pieces.append(b" ".join((self.channels[channel_key].pigs_function,
                                               bytes(str(self.channels[channel_key].pin_bcm), 'utf-8'),
                                               bytes(str(self.channels[channel_key]._clean_value(color_value)), 'utf-8'))))
            string_pieces.append(bytes("{} {}".format(wait_fn, duration), 'utf-8'))

        script_str = b" ".join(string_pieces)

        if repeat:
            try:
                repeat = int(repeat)
            except:
                ValueError('Repeat must be coerceable to an integer, got {}'.format(repeat))

            script_str = b" ".join(("LD v0", str(repeat-1), # "LD (load) variable 0 with number of repeats"
                                    "tag 999",            # create a tag that can be returned to
                                    script_str,           # do the script one time
                                    "dcr v0"              # decrement v0. 
                                    "jp 999")             # jump to the tag at the beginning if v0 is still positive
                                   )

        # turn off when finished
        if finish_off:
            #script_str += b" "
            finish_pieces = [script_str]
            for channel in self.channels.values():
                finish_pieces.append(b" ".join((channel.pigs_function, bytes(str(channel.pin_bcm), 'utf-8'), bytes(str(channel.off), 'utf-8'))))
            script_str = b" ".join(finish_pieces)

        return script_str

    def flash(self, duration, frequency=10, colors=((1,1,1),(0,0,0))):
        """
        Specify a color series by total duration and flash frequency.

        Largely a convenience function for on/off flashes.

        Args:
            duration (int, float): Duration of flash in ms.
            frequency (int, float): Frequency of flashes in Hz
            colors (list): A list of RGB values 0-255 like::

                [[255,255,255],[0,0,0]]

        """
        # Duration is total in ms, frequency in Hz
        # Get number of flashes in duration rounded down

        n_rep = int(float(duration) / 1000. * float(frequency))
        flashes = colors * n_rep

        # Invert frequency to duration for single flash
        # divide by 2 b/c each 'color' is half the duration
        single_dur = ((1. / frequency) * 1000) / 2.

        flash_params = (flashes, single_dur)

        new_flash = True
        if self.flash_params:
            if self.flash_params == flash_params:
                # do nothing
                new_flash = False

        if new_flash:
            self.store_series('flash',
                              colors=flashes,
                              durations=single_dur)
            self.flash_params = flash_params


        self.series('flash')

    #
    # def __getattr__(self, name):
    #     """if a method is called that we don't explicitly redefine, try to apply it to each of our channels"""
    #
    #     attrs = {color: object.__getattribute__(chan, name) for color, chan in self.channels.items()}
    #     if hasattr(list(attrs.values())[0], '__call__'):
    #         def newfunc(*args, **kwargs):
    #             result = {}
    #             # if called with a dictionary of {'r':params, 'g':params, ...}, call that explicitly
    #             if isinstance(args[0], dict):
    #                 if any([k in args[0].keys() for k in self.channels.keys()]):
    #                     for color, method in attrs.items():
    #                         if color in args[0].keys():
    #                             result[color] = method(**args[0][color])
    #                     return result
    #
    #             # otherwise, call all channels with the same argument
    #             for color, method in attrs.items():
    #                 result[color] = method(*args, **kwargs)
    #             return result
    #
    #         return newfunc
    #     else:
    #         return attrs

    def release(self):
        """
        Release each channel and stop pig without calling superclass.
        """
        for chan in self.channels.values():
            chan.release()
        self.pig.stop()


    @property
    def pin(self):
        """
        Dict of the board pin number of each channel, ``{'r' : self.channels['r'].pin, ... }``
        """
        try:
            return {'r':self.channels['r'].pin,
                    'g':self.channels['g'].pin,
                    'b':self.channels['b'].pin}
        except KeyError:
            return {}

    @pin.setter
    def pin(self, pin):
        """
        Pins don't get set like this for the LED_RGB, ignore.
        """
        if pin is None:
            # just remnants of the attempt to set from the GPIO metaclass
            return
        self.logger.warning('pin cant be set via the attribute')

    @property
    def pin_bcm(self):
        """
        Dict of the broadcom pin number of each channel, ``{'r' : self.channels['r'].pin_bcm, ... }``
        """
        try:
            return {'r':self.channels['r'].pin_bcm,
                    'g':self.channels['g'].pin_bcm,
                    'b':self.channels['b'].pin_bcm}
        except KeyError:
            return {}

    @pin_bcm.setter
    def pin_bcm(self, pin_bcm):
        """
        Pins don't get set like this for the LED_RGB, ignore\
        """
        if pin_bcm is None:
            return

        self.logger.warning('pin_bcm cant be set via the attribute')

    @property
    def pull(self):
        return None

    @pull.setter
    def pull(self, direction):
        if direction is None:
            return
        self.logger.warning('pull cant be set via the attribute')


class Solenoid(Digital_Out):
    """
    Solenoid valve for water delivery.

    Args:
        pin (int): Board pin number, converted to BCM on init.
        polarity (0, 1): Whether HIGH opens the port (1) or closes it (0)
        duration (int, float): duration of open, ms.
        vol (int, float): desired volume of reward in uL, must have computed calibration results, see :meth:`~.Terminal.calibrate_ports`
        **kwargs: passed to :class:`Digital_Out`

    Only NC solenoids should be used, as there is no way to guarantee
    that a pin will maintain its voltage when it is released, and you will
    spill water all over the place.

    Attributes:
        calibration (dict): Dict with with line coefficients fitting volume to open duration, see :meth:`~.Terminal.calibrate_ports`.
            Retrieved from prefs, specifically ``prefs.get('PORT_CALIBRATION')[name]``
        mode ('DURATION', 'VOLUME'): Whether open duration is given in ms, or computed from calibration
        duration (int, float): Duration of valve opening, in ms. When set, creates a script 'open' that is used to open the valve for a precise amount of time
    """

    output = True
    type = "SOLENOID"
    DURATION_MIN = 2 #: Minimum allowed duration in ms


    def __init__(self, pin, polarity=1, duration=20, vol=None, **kwargs):
        """

        """
        super(Solenoid, self).__init__(pin, polarity=polarity, **kwargs)
        self.calibration = None
        self._duration = None

        # Pigpio has us create waves to deliver timed output
        # Since we typically only use one duration,
        # we make the wave once and only make it again when asked to
        # We start with passed or default duration (ms)
        if vol:
            self.duration = self.dur_from_vol(vol)
            self.mode = 'VOLUME'

        else:
            self.duration = duration
            self.mode='DURATION'

    @property
    def duration(self):
        return self._duration

    @duration.setter
    def duration(self, duration):
        duration = round(duration)
        if duration < self.DURATION_MIN:
            Warning('Duration cannot be less than {}, so value was clipped. given {}'.format(self.DURATION_MIN, duration))
            duration = self.DURATION_MIN

        self._duration = duration
        self.store_series(id="open", values=self.on, durations=self._duration, unit="ms", finish_off=True, repeat=None)



    def dur_from_vol(self, vol):
        """
        Given a desired volume, compute an open duration.

        Must have calibration available in prefs, see :meth:`~.Terminal.calibrate_ports`.

        Args:
            vol (float, int): desired reward volume in uL

        Returns:
            int: computed opening duration for given volume
        """
        # find our pin name
        if not self.name:
            self.name = self.get_name()

        # legacy, compatibility code -- if calibrations were made in the olde way
        # then try and load them. they will be saved in the new way by the calibration setter in the root class
        # prefs should have loaded any calibration

        if self.calibration is None:

            try:
                self.calibration = prefs.get('PORT_CALIBRATION')[self.name]
            except KeyError:
                # try using name prepended with PORTS_, which happens for hardware objects with implicit names
                self.calibration = prefs.get('PORT_CALIBRATION')[self.name.replace('PORTS_', '')]
            except Exception as e:
                self.logger.exception(f'couldnt get calibration, using default LUT y = 3.5 + 2. got error {e}')
                self.calibration = {'slope': 3.5, 'intercept': 2}
        # compute duration from slope and intercept
        duration = round(float(self.calibration['intercept']) + (float(self.calibration['slope']) * float(vol)))

        return duration


    def open(self, duration=None):
        """
        Open the valve.

        Uses the 'open' script created when assigning duration.

        Args:
            duration (float): If provided, open for this duration instead of
                the duration stored on instantiation.
        """
        if duration:
            self.duration = duration

        self.series(id="open")





"""
Run scripts to setup system dependencies and autopilot plugins

::

    > # to list scripts
    > python3 -m autopilot.setup.run_script --list

    > # to execute one script (setup hifiberry soundcard)
    > python3 -m autopilot.setup.run_script hifiberry

    > # to execute multiple scripts
    > python3 -m autopilot.setup.run_script hifiberry jackd

"""
import typing
import subprocess
from autopilot.setup.scripts import SCRIPTS
import argparse
import sys

parser = argparse.ArgumentParser(description="Run autopilot setup script(s)")
parser.add_argument('scripts', nargs='*', type=str)
parser.add_argument('--list', help="list available setup scripts!", action='store_true')


def call_series(commands, series_name=None):
    """
    Call a series of commands, giving a single return code on completion or failure

    :param commands:
    :return:
    """
    if series_name:
        print('\n\033[1;37;42m Running commands for {}\u001b[0m'.format(series_name))

    # have to just combine them -- can't do multiple calls b/c shell doesn't preserve between them
    combined_calls = ""
    last_command = len(commands)-1
    for i, command in enumerate(commands):
        join_with = " && "

        if isinstance(command, str):
            # just a command, default necessary
            combined_calls += command
        elif isinstance(command, dict):
            combined_calls += command['command']

            if command.get('optional', False):
                join_with = "; "

        if i < last_command:
            combined_calls += join_with


    print('Executing:\n    {}'.format(combined_calls))

    result = subprocess.run(combined_calls, shell=True, executable='/bin/bash')

    status = False
    if result.returncode == 0:
        status = True

    if series_name:
        if status:
            print('\n\033[1;37;42m  {} Successful, you lucky duck\u001b[0m'.format(series_name))
        else:
            print('\n\033[1;37;41m  {} Failed, check the error message & ur crystal ball\u001b[0m'.format(series_name))

    return status


def run_script(script_name):
    if script_name in SCRIPTS.keys():
        call_series(SCRIPTS[script_name]['commands'], script_name)
    else:
        raise NameError('No script named {}, must be one of {}'.format(script_name, "\n".join(SCRIPTS.keys())))

def run_scripts(scripts, return_all=False,print_status=True) -> typing.Union[bool, typing.Dict[str,bool]]:
    """
    Run a series of scripts, printing results

    Args:
        scripts (list): list of script names

    Returns:
        bool: success or failure of scripts - ``True`` if all were successful, ``False`` otherwise.
    """
    env_results = {}
    for script_name in scripts:
        commands = SCRIPTS[script_name].get('commands', None)
        if commands is not None:
            env_results[script_name] = call_series(commands, script_name)

    # indicate global success
    success = True

    # make results string
    if print_status:
        env_result = "\033[0;32;40m\n--------------------------------\nScript Results:\n"
        for config, result in env_results.items():
            if result:
                env_result += "  [ SUCCESS ] "
            else:
                env_result += "  [ FAILURE ] "
                success = False

            env_result += config
            env_result += '\n'

        env_result += '--------------------------------\u001b[0m'

        print(env_result)

    if return_all:
        return env_results
    else:
        return success


def list_scripts():
    print('\nAvailable Scripts:\n----------------------------')

    # find longest script name
    longest_name = 0
    for script_name in SCRIPTS.keys():
        if len(script_name) > longest_name:
            longest_name = len(script_name)


    for script_name in sorted(SCRIPTS.keys()):
        pad = " " * (longest_name - len(script_name))
        print(f'\033[1;37;42m{script_name}{pad}\u001b[0m : {SCRIPTS[script_name]["text"]}')

if __name__ == "__main__":
    args = parser.parse_args()

    if args.list:
        list_scripts()
        sys.exit()
    elif args.scripts:
        res = run_scripts(args.scripts)
        if res:
            sys.exit()
        else:
            sys.exit(1)
    else:
        raise RuntimeError('Need to give name of one or multiple scripts, ')



import datetime
import itertools
import tables
import threading
import typing

import autopilot
from autopilot.tasks import Task
from autopilot.stim import init_manager
from collections import OrderedDict as odict

# This declaration allows Subject to identify which class in this file contains the task class. Could also be done with __init__ but yno I didnt for no reason.
# TODO: Move this to __init__
TASK = 'Nafc'

class Nafc(Task):
    """
    A Two-alternative forced choice task.

    *(can't have number as first character of class.)*

    **Stages**

    * **request** - compute stimulus, set request trigger in center port.
    * **discrim** - respond to input, set reward/punishment triggers on target/distractor ports
    * **reinforcement** - deliver reward/punishment, end trial.

    Attributes:
        target ("L", "R"): Correct response
        distractor ("L", "R"): Incorrect response
        stim : Current stimulus
        response ("L", "R"): Response to discriminand
        correct (0, 1): Current trial was correct/incorrect
        correction_trial (bool): If using correction trials, last trial was a correction trial
        trial_counter (:class:`itertools.count`): Which trial are we on?
        discrim_playing (bool): Is the stimulus playing?
        bailed (0, 1): Subject answered before stimulus was finished playing.
        current_stage (int): As each stage is reached, update for asynchronous event reference

    """
    STAGE_NAMES = ["request", "discrim", "reinforcement"]

    # Class attributes


    # List of needed params, returned data and data format.
    # Params are [name]={'tag': Human Readable Tag, 'type': 'int', 'float', 'bool', etc.}
    PARAMS = odict()
    # TODO: Reward no longer just duration -- fix with parameter structure
    PARAMS['reward']         = {'tag':'Reward Duration (ms)',
                                'type':'int'}
    PARAMS['req_reward']     = {'tag':'Request Rewards',
                                'type':'bool'}
    PARAMS['punish_stim']   = {'tag':'White Noise Punishment',
                                'type':'bool'}
    PARAMS['punish_dur']     = {'tag':'Punishment Duration (ms)',
                                'type':'int'}
    PARAMS['correction']     = {'tag':'Correction Trials',
                                'type':'bool'}
    PARAMS['correction_pct'] = {'tag':'% Correction Trials',
                                'type':'int',
                                'depends':{'correction':True}}
    PARAMS['bias_mode']      = {'tag':'Bias Correction Mode',
                                'type':'list',
                                'values':{'None':0, 'Proportional':1, 'Thresholded Proportional':2}}
    PARAMS['bias_threshold'] = {'tag': 'Bias Correction Threshold (%)',
                                'type':'int',
                                'depends':{'bias_mode':2}}
    #PARAMS['timeout']        = {'tag':'Delay Timeout (ms)',
    #                            'type':'int'}
    PARAMS['stim']           = {'tag':'Sounds',
                                'type':'sounds'}

    # Set plot params, which data should be plotted, its default shape, etc.
    # TODO: Plots should take the default type, but options panel should be able to set - eg. corrects are done by rolling mean as default, but can be made points
    PLOT = {
        'data': {
            'target'   : 'point',
            'response' : 'segment',
            'correct'  : 'rollmean'
        },
        'chance_bar'  : True, # Draw a red bar at 50%
        'roll_window' : 50 # number of trials to roll window over
    }

    # PyTables Data descriptor
    # for numpy data types see http://docs.scipy.org/doc/numpy/reference/arrays.dtypes.html#arrays-dtypes-constructing
    class TrialData(tables.IsDescription):
        # This class allows the Subject object to make a data table with the correct data types. You must update it for any new data you'd like to store
        trial_num = tables.Int32Col()
        target = tables.StringCol(1)
        response = tables.StringCol(1)
        correct = tables.Int32Col()
        correction = tables.Int32Col()
        RQ_timestamp = tables.StringCol(26)
        DC_timestamp = tables.StringCol(26)
        bailed = tables.Int32Col()

    HARDWARE = {
        'POKES':{
            'L': "Digital_In",
            'C': "Digital_In",
            'R': "Digital_In"
        },
        'LEDS':{
            # TODO: use LEDs, RGB vs. white LED option in init
            'L': "LED_RGB",
            'C': "LED_RGB",
            'R': "LED_RGB"
        },
        'PORTS':{
            'L': "Solenoid",
            'C': "Solenoid",
            'R': "Solenoid"
        }
    }


    def __init__(self, stage_block=None, stim=None, reward=50, req_reward=False,
                 punish_stim=False, punish_dur=100, correction=False, correction_pct=50.,
                 bias_mode=False, bias_threshold=20, stim_light=True, **kwargs):
        """
        Args:
            stage_block (:class:`threading.Event`): Signal when task stages complete.
            stim (dict): Stimuli like::


                "sounds": {
                    "L": [{"type": "Tone", ...}],
                    "R": [{"type": "Tone", ...}]
                }

            reward (float): duration of solenoid open in ms
            req_reward (bool): Whether to give a water reward in the center port for requesting trials
            punish_stim (bool): Do a white noise punishment stimulus
            punish_dur (float): Duration of white noise in ms
            correction (bool): Should we do correction trials?
            correction_pct (float):  (0-1), What proportion of trials should randomly be correction trials?
            bias_mode (False, "thresholded_linear"): False, or some bias correction type (see :class:`.managers.Bias_Correction` )
            bias_threshold (float): If using a bias correction mode, what threshold should bias be corrected for?
            current_trial (int): If starting at nonzero trial number, which?
            stim_light (bool): Should the LED be turned blue while the stimulus is playing?
            **kwargs:
        """
        super(Nafc, self).__init__()

        # Fixed parameters
        # Because the current protocol is json.loads from a string,
        # we should explicitly type everything to be safe.
        if isinstance(reward, dict):
            self.reward = reward
        else:
            self.reward         = {'type':'duration',
                                   'value': float(reward)}
        self.req_reward     = bool(req_reward)
        self.punish_stim   = bool(punish_stim)
        self.punish_dur     = float(punish_dur)
        self.correction     = bool(correction)
        self.correction_pct = float(correction_pct)/100
        self.bias_mode      = bias_mode
        self.bias_threshold = float(bias_threshold)/100
        self.stim_light      = bool(stim_light)
        #self.timeout        = int(timeout)

        # Variable Parameters
        self.target = None
        self.distractor = None
        self.stim = None
        self.response = None
        self.correct = None
        self.correction_trial = False
        #self.discrim_finished = False # Set to true once the discrim stim has finished, used for punishing leaving C early
        self.discrim_playing = False
        self.current_stage = None # Keep track of stages so some asynchronous callbacks know when it's their turn
        self.bailed = 0

        # We make a list of the variables that need to be reset each trial so it's easier to do so
        self.resetting_variables = [self.response, self.bailed]

        # This allows us to cycle through the task by just repeatedly calling self.stages.next()
        stage_list = [self.request, self.discrim, self.reinforcement]
        self.num_stages = len(stage_list)
        self.stages = itertools.cycle(stage_list)

        # Initialize hardware
        self.init_hardware()
        self.logger.debug('Hardware initialized')

        # Set reward values for solenoids
        # TODO: Super inelegant, implement better with reward manager
        if self.reward['type'] == "volume":
            self.set_reward(vol=self.reward['value'])
        else:
            self.set_reward(duration=self.reward['value'])

        # Initialize stim manager
        if not stim:
            raise RuntimeError("Cant instantiate task without stimuli!")
        else:
            self.stim_manager = init_manager(stim)

        self.logger.debug('stimuli initialized')

        # give the sounds a function to call when they end
        self.stim_manager.set_triggers(self.stim_end)
        self.logger.debug('stim triggers set')

        if self.correction:
            self.stim_manager.do_correction(self.correction_pct)
            self.logger.debug(f'correction trials initialized, correction_pct: {self.correction_pct}')

        if bias_mode:
            self.stim_manager.do_bias(mode=self.bias_mode,
                                      thresh=self.bias_threshold)
            self.logger.debug(f'bias correction initialized, bias_mode: {self.bias_mode}, threshold: {self.bias_threshold}')
        self.logger.debug('Stimulus manager initialized')

        # If we aren't passed an event handler
        # (used to signal that a trigger has been tripped),
        # we should warn whoever called us that things could get a little screwy
        if not stage_block:
            raise Warning('No stage_block Event() was passed, youll need to handle stage progression on your own')
        else:
            self.stage_block = stage_block

        self.logger.debug('finished initializing Nafc class')

    #
    # def center_out(self, pin, level, tick):
    #     """
    #
    #     """
    #     # Called when something leaves the center pin,
    #     # We use this to handle the subject leaving the port early
    #     if self.discrim_playing:
    #         self.bail_trial()


    ##################################################################################
    # Stage Functions
    ##################################################################################
    def request(self,*args,**kwargs):
        """
        Stage 0: compute stimulus, set request trigger in center port.

        Returns:
            data (dict): With fields::

                {
                'target': self.target,
                'trial_num' : self.current_trial,
                'correction': self.correction_trial,
                'type': stimulus type,
                **stim.PARAMS
                }

        """
        # Set the event lock
        self.stage_block.clear()

        # Reset all the variables that need to be
        for v in self.resetting_variables:
            v = None

        # reset triggers if there are any left
        self.triggers = {}

        # get next stim
        self.target, self.distractor, self.stim = self.stim_manager.next_stim()
        # buffer it
        self.stim.buffer()

        # if we're doing correction trials, check if this is one
        if self.correction:
            self.correction_trial = self.stim_manager.correction_trial

        # Set sound trigger and LEDs
        self.triggers['C'] = [self.stim.play, self.stim_start]
        if self.stim_light:
            change_to_blue = lambda: self.hardware['LEDS']['C'].set([0, 0, 255])
            self.triggers['C'].append(change_to_blue)
        else:
            # just turn the center light off and side lights on immediately.
            turn_off = lambda: self.set_leds({
                'L': [0,255,0],
                'R': [0,255,0]
            })
            #turn_off = lambda: self.hardware['LEDS']['C'].set([0,0,0])
            self.triggers['C'].append(turn_off)

        if self.req_reward:
            self.triggers['C'].append(self.hardware['PORTS']['C'].open)


        self.current_trial = next(self.trial_counter)
        data = {
            'target':self.target,
            'trial_num' : self.current_trial,
            'correction':self.correction_trial
        }
        # get stim info and add to data dict
        sound_info = {k:getattr(self.stim, k) for k in self.stim.PARAMS}
        data.update(sound_info)
        data.update({'type':self.stim.type})

        self.current_stage = 0

        # wait on punish block
        # FIXME: Only waiting to test whether this is where the bug that hangs after this stage is
        self.punish_block.wait(20)

        # set to green in the meantime
        self.set_leds({'C': [0, 255, 0]})


        return data

    def discrim(self,*args,**kwargs):
        """
        Stage 1:  respond to input, set reward/punishment triggers on target/distractor ports

        Returns:
            data (dict): With fields::
                {
                'RQ_timestamp': datetime.datetime.now().isoformat(),
                'trial_num': self.current_trial,
                }

        """
        # moust just poked in center, set response triggers
        self.stage_block.clear()

        self.triggers[self.target] = [lambda: self.respond(self.target), self.hardware['PORTS'][self.target].open]
        self.triggers[self.distractor] = [lambda: self.respond(self.distractor), self.punish]

        # TODO: Handle timeout

        # Only data is the timestamp
        data = {'RQ_timestamp': datetime.datetime.now().isoformat(),
                'trial_num': self.current_trial}
        self.current_stage = 1
        return data

    def reinforcement(self,*args,**kwargs):
        """
        Stage 2 - deliver reward/punishment, end trial.

        Returns:
            data (dict): With fields::

                 {
                'DC_timestamp': datetime.datetime.now().isoformat(),
                'response': self.response,
                'correct': self.correct,
                'bailed': self.bailed,
                'trial_num': self.current_trial,
                'TRIAL_END': True
                }
        """
        # We do NOT clear the task event flag here because we want
        # the pi to call the next stage immediately
        # We are just filling in the last data
        # and performing any calculations we need for the next trial
        if self.bailed:
            self.bailed = 0
            data = {
                'DC_timestamp': datetime.datetime.now().isoformat(),
                'bailed':1,
                'trial_num': self.current_trial,
                'TRIAL_END':True
            }
            return data

        if self.response == self.target:
            self.correct = 1
        else:
            self.correct = 0

        # update stim manager
        self.stim_manager.update(self.response, self.correct)


        data = {
            'DC_timestamp': datetime.datetime.now().isoformat(),
            'response':self.response,
            'correct':self.correct,
            'bailed':0,
            'trial_num': self.current_trial,
            'TRIAL_END':True
        }
        self.current_stage = 2
        return data

    def punish(self):
        """
        Flash lights, play punishment sound if set
        """
        # TODO: If we're not in the last stage (eg. we were timed out after stim presentation), reset stages
        self.punish_block.clear()

        if self.punish_stim:
            self.stim_manager.play_punishment()

        # self.set_leds()
        self.flash_leds()
        threading.Timer(self.punish_dur / 1000., self.punish_block.set).start()


    def respond(self, pin):
        """
        Set self.response

        Args:
            pin: Pin to set response to
        """
        self.response = pin

    def stim_start(self):
        """
        mark discrim_playing = true
        """
        self.discrim_playing = True


    def stim_end(self):
        """
        called by stimulus callback

        set outside lights blue
        """
        # Called by the discrim sound's table trigger when playback is finished
        # Used in punishing leaving early
        self.discrim_playing = False
        #if not self.bailed and self.current_stage == 1:
        if self.stim_light:
            self.set_leds({'L':[0,255,0], 'R':[0,255,0]})

    # def bail_trial(self):
    #     # If a timer ends or the subject pulls out too soon, we punish and bail
    #     self.bailed = 1
    #     self.triggers = {}
    #     self.punish()
    #     self.stage_block.set()

    # def clear_triggers(self):
    #     for pin in self.hardware.values():
    #         pin.clear_cb()

    def flash_leds(self):
        """
        flash lights for punish_dir
        """
        for k, v in self.hardware['LEDS'].items():
            if isinstance(v, autopilot.get_hardware('LED_RGB')):
                v.flash(self.punish_dur)



"""
Hardware that uses USB
"""
import sys
import threading
import time

from queue import Queue, Empty

import numpy as np
from inputs import devices

from autopilot import prefs
from autopilot.networking import Net_Node
from autopilot.hardware import Hardware


class Wheel(Hardware):
    """
    A continuously measured mouse wheel.

    Uses a USB computer mouse.

    Warning:
        'vel' thresh_type not implemented
    """

    input   = True
    type    = "Wheel"
    trigger = False # even though this is a triggerable option, typically don't want to assign a cb and instead us a GPIO
    # TODO: Make the standard-style trigger.
    # TODO: Make wheel movements available locally with a deque

    THRESH_TYPES = ['dist', 'x', 'y', 'vel']

    MODES = ('vel_total', 'steady', 'dist', 'timed')

    MOVE_DTYPE = [('vel', 'i4'), ('dir', 'U5'), ('timestamp', 'f8')]

    def __init__(self, mouse_idx=0, fs=10, thresh=100, thresh_type='dist', start=True,
                 digi_out = False, mode='vel_total', integrate_dur=5):
        """
        Args:
            mouse_idx (int):
            fs (int):
            thresh (int):
            thresh_type ('dist'):
            start (bool):
            digi_out (:class:`~.Digital_Out`, bool):
            mode ('vel_total'):
            integrate_dur (int):
        """

        # try to get mouse from inputs
        # TODO: More robust - specify mouse by hardware attrs
        try:
            self.mouse = devices.mice[mouse_idx]
        except IndexError:
            Warning('Could not find requested mouse with index {}\nAttempting to use mouse idx 0'.format(mouse_idx))
            self.mouse = devices.mice[0]

        # frequency of our updating
        self.fs = fs
        # time between updates
        self.update_dur = 1./float(self.fs)

        self.thresh = thresh
        # thresh type can be 'dist', 'x', 'y', or 'vel'
        if thresh_type not in self.THRESH_TYPES:
            ValueError('thresh_type must be one of {}, given {}'.format(self.THRESH_TYPES, thresh_type))
        self.thresh_type = thresh_type

        # mode can be 'vel_total', 'vel_x', 'vel_y' or 'dist' - report either velocity or distance
        # mode can also be '
        # TODO: Do two parameters - type 'vel' or 'dist' and measure 'x', 'y', 'total'z
        self.mode = mode
        # TODO: Implement this

        if self.mode == "steady":
            self.thresh_val = np.array([(0, "REL_Y", 0)], dtype=self.MOVE_DTYPE)
        else:
            self.thresh_val = 0.0

        self.integrate_dur = integrate_dur


        # event to signal quitting
        self.quit_evt = threading.Event()
        self.quit_evt.clear()
        # event to signal when to start accumulating movements to trigger
        self.measure_evt = threading.Event()
        self.measure_time = 0
        # queue to I/O mouse movements summarized at fs Hz
        self.q = Queue()
        # lock to prevent race between putting and getting
        self.qlock = threading.Lock()

        self.listens = {'MEASURE':self.l_measure,
                        'CLEAR':self.l_clear,
                        'STOP':self.l_stop}
        self.node = Net_Node('wheel_{}'.format(mouse_idx),
                             upstream=prefs.get('NAME'),
                             port=prefs.get('MSGPORT'),
                             listens=self.listens,
                             )

        # if we are being used in a child object, we send our trigger via a GPIO pin
        self.digi_out = digi_out



        self.thread = None

        if start:
            self.start()


    def start(self):
        self.thread = threading.Thread(target=self._record)
        self.thread.daemon = True
        self.thread.start()

    def _mouse(self):
        while self.quit_evt:
            events = self.mouse.read()
            self.q.put(events)

    def _record(self):
        moves = np.array([], dtype=self.MOVE_DTYPE)

        threading.Thread(target=self._mouse).start()

        last_update = time.time()

        while not self.quit_evt.is_set():

            try:
                events = self.q.get_nowait()
            except Empty:
                events = None

            if events is None:
                move = np.array([(0, "REL_Y", 0)], dtype=self.MOVE_DTYPE)
            else:
                # make a numpy record array of events with 3 fields:
                # velocity, dir(ection), timestamp (system seconds)
                move = np.array([(int(event.state), event.code, float(event.timestamp))\
                                 for event in events if event.code in ('REL_X', 'REL_Y')],
                                dtype=self.MOVE_DTYPE)
            moves = np.concatenate([moves, move])

            # If we have been told to start measuring for a trigger...
            if self.measure_evt.is_set():
                do_trigger = self.check_thresh(move)
                if do_trigger:
                    self.thresh_trig()
                    self.measure_evt.clear()
                # take the integral of velocities



            # If it's time to report velocity, do it.
            nowtime = time.time()
            if (nowtime-last_update)>self.update_dur:

                # TODO: Implement distance/position reporting
                y_vel = self.calc_move(moves, 'y')
                x_vel = self.calc_move(moves, 'x')

                self.node.send(key='CONTINUOUS', value={'x':x_vel, 'y':y_vel, 't':nowtime},
                               repeat=False)

                moves = np.array([], dtype=self.MOVE_DTYPE)

                last_update = nowtime

    def check_thresh(self, move):
        """
        Updates thresh_val and checks whether it's above/below threshold

        Args:
            move (np.array): Structured array with fields ('vel', 'dir', 'timestamp')

        Returns:

        """

        # Determine whether the threshold was surpassed
        do_trigger = False
        if self.mode == 'vel_total':
            thresh_update = self.calc_move(move)
            # If instantaneous velocity is above thresh...
            if thresh_update > self.thresh:
                do_trigger = True

        elif self.mode == 'steady':
            # If movements in the recent past are below a certain value
            # self.thresh_val should be set to a structured array by l_measure
            try:
                self.thresh_val = np.concatenate([self.thresh_val, move])
            except TypeError:
                print('THRESH_VAL:', self.thresh_val, 'MOVE:', move)
            # trim to movements in the time window
            thresh_val = self.thresh_val[self.thresh_val['timestamp'] > time.time()-self.integrate_dur]

            thresh_update = self.calc_move(thresh_val)

            if (thresh_update < self.thresh) and (self.measure_time+self.integrate_dur < time.time()):
                do_trigger = True

        elif self.mode == 'dist':
            thresh_update = self.calc_move(move)
            self.thresh_val += thresh_update

            if self.thresh_val > self.thresh:
                do_trigger = True

        else:
            Warning ("mode is not defined! mode is {}".format(self.mode))


        return do_trigger

    def calc_move(self, move, thresh_type=None):
        """
        Calculate distance move depending on type (x, y, total dist)

        Args:
            move ():
            thresh_type ():

        Returns:

        """

        if thresh_type is None:
            thresh_type = self.thresh_type

        # FIXME: rly inefficient
        # get the value of the movement depending on what we're measuring
        if thresh_type == 'x':

            distance = np.sum(move['vel'][move['dir'] == "REL_X"])
        elif thresh_type == 'y':
            distance = np.sum(move['vel'][move['dir'] == "REL_Y"])
        elif thresh_type == "dist":
            x_dist = np.sum(move['vel'][move['dir'] == "REL_X"])
            y_dist = np.sum(move['vel'][move['dir'] == "REL_Y"])
            distance = np.abs(np.sqrt(float(x_dist ** 2) + float(y_dist ** 2)))

        return distance

    def thresh_trig(self):


        if self.digi_out:
            self.digi_out.pulse()

        self.measure_evt.clear()




    def assign_cb(self, trigger_fn):
        # want to have callback write an output pin -- so callback should go back to
        # the task to write a GPIO pin.
        self.trig_fn = trigger_fn

    def l_measure(self, value):
        """
        Task has signaled that we need to start measuring movements for a trigger


        Args:
            value ():
        """

        if 'mode' in value.keys():
            if value['mode'] in self.MODES:
                self.mode = value['mode']
            else:
                Warning('incorrect mode sent: {}, needs to be one of {}'.format(value['mode'], self.MODES))

        if 'thresh' in value.keys():
            self.thresh = float(value['thresh'])

        if self.mode == "steady":
            self.thresh_val = np.array([(0, "REL_Y", 0)], dtype=self.MOVE_DTYPE)
        else:
            self.thresh_val = 0.0
        self.measure_time = time.time()

        self.measure_evt.set()

        sys.stdout.flush()

    def l_clear(self, value):
        """
        Stop measuring!

        Args:
            value ():

        Returns:

        """
        self.measure_evt.clear()

    def l_stop(self, value):
        """
        Stop measuring and clear system resources
        Args:
            value ():

        Returns:

        """

        self.measure_evt.set()
        self.release()

    def release(self):
        self.quit_evt.clear()


class Scale(Hardware):
    """
    Note:
        Not implemented, working on using a digital scale to
        make weighing faster.
    """
    MODEL={
        'stamps.com':{
            'vendor_id':0x1446,
            'product_id': 0x6a73

        }
    }
    def __init__(self, model='stamps.com', vendor_id = None, product_id = None):
        """
        Args:
            model:
            vendor_id:
            product_id:
        """
        self.vendor_id = self.MODEL[model]['vendor_id']
        self.product_id = self.MODEL[model]['product_id']

        if vendor_id:
            self.vendor_id = vendor_id
        if product_id:
            self.product_id = product_id

        # find device
        self.device = usb.core.find(idVendor=self.vendor_id,
                                    idProduct=self.product_id)
        # default configuration
        self.device.set_configuration()